#!/usr/bin/env python

# pmx  Copyright Notice
# ============================
#
# The pmx source code is copyrighted, but you can freely use and
# copy it as long as you don't change or remove any of the copyright
# notices.
#
# ----------------------------------------------------------------------
# pmx is Copyright (C) 2006-2013 by Daniel Seeliger
#
#                        All Rights Reserved
#
# Permission to use, copy, modify, distribute, and distribute modified
# versions of this software and its documentation for any purpose and
# without fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both the copyright notice and
# this permission notice appear in supporting documentation, and that
# the name of Daniel Seeliger not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# DANIEL SEELIGER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL DANIEL SEELIGER BE LIABLE FOR ANY
# SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ----------------------------------------------------------------------

"""jobscript generation
"""

import os
import sys
import types
import re
import logging
from glob import glob

class Jobscript:
    """Class for jobscript generation

    Parameters
    ----------
    ...

    Attributes
    ----------
    ....

    """

    def __init__(self, **kwargs):

        self.queue = 'SGE' # could be SLURM
        self.simtime = 24 # hours
        self.simcpu = 2 # CPU default
        self.bGPU = True
        self.fname = 'jobscript'
        self.jobname = 'jobName'
        self.modules = []
        self.source = []
        self.export = []
        self.cmds = [] # commands to add to jobscript
        self.gmx = None
        self.header = ''
        self.cmdline = ''
        self.partition = ''
        
        for key, val in kwargs.items():
            setattr(self,key,val)                 
             
    def create_jobscript( self ):
        # header
        self._create_header()
        # commands
        self._create_cmdline()
        # write
        self._write_jobscript()    
        
    def _write_jobscript( self ):
        fp = open(self.fname,'w')
        fp.write(self.header)
        fp.write(self.cmdline)
        fp.close()
            
    def _add_to_jobscriptFile( self ):
        fp = open(self.fname,'a')
        fp.write('{0}\n'.format(cmd))
        fp.close()           
            
    def _create_cmdline( self ):
        if isinstance(self.cmds,list)==True:
            for cmd in self.cmds:
                self.cmdline = '{0}{1}\n'.format(self.cmdline,cmd)
        else:
            self.cmdline = cmds
            
    def _create_header( self ):
        moduleline = ''
        sourceline = ''
        exportline = ''
        partitionline = self.partition
        for m in self.modules:
            moduleline = '{0}\nmodule load {1}'.format(moduleline,m)
        for s in self.source:
            sourceline = '{0}\nsource {1}'.format(sourceline,s)
        for e in self.export:
            exportline = '{0}\export load {1}'.format(exportline,e)
        gpuline = ''
        if self.bGPU==True:
            if self.queue == 'SGE':
                gpuline = '#$ -l gpu=1'
            elif self.queue == 'SLURM':
                gpuline = '#SBATCH --gres=gpu:1'
        gmxline = ''
        if self.gmx!=None:
            gmxline = 'export GMXRUN="{gmx} -ntomp {simcpu} -ntmpi 1"'.format(gmx=self.gmx,simcpu=self.simcpu)            
            
        if self.queue=='SGE':
            self._create_SGE_header(moduleline,sourceline,exportline,gpuline,gmxline,partitionline)
        elif self.queue=='SLURM':
            self._create_SLURM_header(moduleline,sourceline,exportline,gpuline,gmxline,partitionline)
        
    def _create_SGE_header( self,moduleline,sourceline,exportline,gpuline,gmxline, partitionline ):    
        self.header = '''#$ -S /bin/bash
#$ -N {jobname}
#$ -l h_rt={simtime}:00:00
#$ -cwd
#$ -pe *_fast {simcpu}
{gpu}

{source}
{modules}
{export}

{gmx}
'''.format(jobname=self.jobname,simcpu=self.simcpu,simtime=self.simtime,gpu=gpuline,
           source=sourceline,modules=moduleline,export=exportline,
           gmx=gmxline)
        
        
    def _create_SLURM_header( self,moduleline,sourceline,exportline,gpuline,gmxline,partitionline):
        fp = open(self.fname,'w')

        # optionally, can create a partition entry
        partition = ''
        if partitionline!=None and partitionline!='':
            partition = "#SBATCH --partition={0}\n".format(partitionline)

        self.header = '''#!/bin/bash
#SBATCH --job-name={jobname}
#SBATCH --get-user-env
#SBATCH -N 1
#SBATCH -n {simcpu}
#SBATCH -t {simtime}:00:00
{partition}
{gpu}

{source}
{modules}
{export}

{gmx}
'''.format(jobname=self.jobname,simcpu=self.simcpu,simtime=self.simtime,partition=partition,
           gpu=gpuline,source=sourceline,modules=moduleline,export=exportline,
           gmx=gmxline)

    def _submission_script( self, jobfolder, counter, simType='eq', frNum=80, bArray=True ):
        fname = '{0}/submit.py'.format(jobfolder)
        fp = open(fname,'w')
        fp.write('import os\n')
        fp.write('for i in range(0,{0}):\n'.format(counter))
        if self.queue=='SGE':
            cmd = '\'qsub jobscript{0}\'.format(i)'
            if ((simType=='ti') or ('transition' in simType)) and (bArray==True):
                cmd = '\'qsub -t 1-'+str(frNum)+':1 jobscript{0}\'.format(i)'
        elif self.queue=='SLURM':
            cmd = '\'sbatch jobscript{0}\'.format(i)'
            if ((simType=='ti') or ('transition' in simType)) and (bArray==True):
                cmd = '\'sbatch --array=1-'+str(frNum)+' jobscript{0}\'.format(i)'
        fp.write('    os.system({0})\n'.format(cmd))
        fp.close()

    



# pmx  Copyright Notice
# ============================
#
# The pmx source code is copyrighted, but you can freely use and
# copy it as long as you don't change or remove any of the copyright
# notices.
#
# ----------------------------------------------------------------------
# pmx is Copyright (C) 2006-2018 by Daniel Seeliger
#
#                        All Rights Reserved
#
# Permission to use, copy, modify, distribute, and distribute modified
# versions of this software and its documentation for any purpose and
# without fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both the copyright notice and
# this permission notice appear in supporting documentation, and that
# the name of Daniel Seeliger not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# DANIEL SEELIGER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL DANIEL SEELIGER BE LIABLE FOR ANY
# SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ----------------------------------------------------------------------

"""**pmx** is a essentially a collection of classes and functions that allow
the manipulation of molecular structure and topolofy files. It has been thought
to be mainly used in conjunction with the Gromacs simulation package. In fact,
**pmx** makes use of some Gromacs functions to read/write trajectory data
and to allow fast neighborsearching from a python script. With time, **pmx**
has also become a library for the setup and analysis of alchemical free energy
calculations in Gromacs."""


from .atom import Atom
from .molecule import Molecule
from .chain import Chain
from .model import Model
from .forcefield import Topology
from .estimators import BAR, Crooks, Jarz
from .analysis import *
from .alchemy import *
from . import gmx

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



#!/usr/bin/env python
# pmx  Copyright Notice
# ============================
#
# The pmx source code is copyrighted, but you can freely use and
# copy it as long as you don't change or remove any of the copyright
# notices.
#
# ----------------------------------------------------------------------
# pmx is Copyright (C) 2006-2013 by Daniel Seeliger
#
#                        All Rights Reserved
#
# Permission to use, copy, modify, distribute, and distribute modified
# versions of this software and its documentation for any purpose and
# without fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both the copyright notice and
# this permission notice appear in supporting documentation, and that
# the name of Daniel Seeliger not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# DANIEL SEELIGER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL DANIEL SEELIGER BE LIABLE FOR ANY
# SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ----------------------------------------------------------------------

import os
import argparse
from pmx.forcefield import Topology
from pmx.utils import ff_selection
from pmx.utils import multiple_replace
from pmx.alchemy import gen_hybrid_top, write_split_top
from pmx.scripts.cli import check_unknown_cmd


def _change_outfile_format(filename, ext):
    head, tail = os.path.split(filename)
    name, ex = os.path.splitext(tail)
    new_name = os.path.join(head, name+'.'+ext)
    return new_name


# =============
# Input Options
# =============
def parse_options():
    parser = argparse.ArgumentParser(description='''
This script fills in the B state to a topology file (itp or top) according to
the hybrid residues present in the file. If you provide a top file with
include statemets, by default the script will run through the included itp
files too; this can turned off using the --norecursive flag.

You need to use this script after having mutated a structure file with pmx
mutate, and after having passed that mutated structure through pdb2gmx.
''')

    parser.add_argument('-p',
                        metavar='topol',
                        dest='intop',
                        type=str,
                        help='Input topology file (itp or top). '
                        'Default is "topol.top"',
                        default='topol.top')
    parser.add_argument('-o',
                        metavar='outfile',
                        dest='outfile',
                        type=str,
                        help='Output topology file. '
                        'Default is "pmxtop.top"',
                        default='pmxtop.top')
    parser.add_argument('-ff',
                        metavar='ff',
                        dest='ff',
                        type=str.lower,
                        help='Force field to use. If -p is a top file, it is '
                        'not necessary to specify the forcefield, as it will '
                        'be determined automatically. If -p is an itp file, '
                        'then -ff is needed, and if not provided a list of '
                        'available ff will be shown.',
                        default=None)
    parser.add_argument('--split',
                        dest='split',
                        help='Write separate topologies for the vdW and charge'
                        ' transformations.',
                        default=False,
                        action='store_true')
    parser.add_argument('--scale_mass',
                        dest='scale_mass',
                        help='Scale the masses of dummy atoms by multiplying '
                        'with scale_mass factor',
                        default=0.33, #TODO: need a function to check for oscillation period: need to ensure 2*pi*sqrt(m1*m2/((m1+m2)*k)) > 10*dt
                        action='store')
    parser.add_argument('--scale_dih',
                        dest='scale_dih',
                        help='Scale the dihedrals that have a dummy. ',
                        default=1.0,
                        action='store')
    parser.add_argument('--norecursive',
                        dest='recursive',
                        help='Whether to fill the B states also for all itp '
                        'files included in the provided topology file. '
                        'Default is True. This flag sets it to False.',
                        default=True,
                        action='store_false')

    args, unknown = parser.parse_known_args()
    check_unknown_cmd(unknown)

    # ff selection is required if file provided is an itp
    if args.intop.split('.')[-1] == 'itp' and args.ff is None:
        args.ff = ff_selection()

    return args


# ====
# Main
# ====
def main(args):

    top_file = args.intop
    top_file_ext = top_file.split('.')[-1]
    outfile = args.outfile
    ff = args.ff
    scale_mass = float(args.scale_mass)
    recursive = args.recursive
    scaleDih = float(args.scale_dih)

    # if input is itp but output is else, rename output
    if top_file_ext == 'itp' and outfile.split('.')[-1] != 'itp':
        outfile = _change_outfile_format(outfile, 'itp')
        print('log_> Setting outfile name to %s' % outfile)

    # load topology file
    topol = Topology(top_file, ff=ff, version='new')

    # fill the B states
    pmxtop, pmxitps = gen_hybrid_top(topol=topol, recursive=recursive,
                                     verbose=True, scaleDih=scaleDih)

    # get the base path for the .itp file output
    basepath = os.path.abspath(outfile).rsplit('/',1)[0]

    # write hybrid itps if present
    replace = {}
    if len(pmxitps) > 0:
        for pmxitp in pmxitps:
            itp_fn = os.path.basename(pmxitp.filename)
            out_fn = '{0}/pmx_{1}'.format(basepath,itp_fn)
            # store old/new itp names for replacement in top file
            replace[itp_fn] = 'pmx_{0}'.format(itp_fn) #out_fn
            print('\nlog_> Writing itp file "%s""' % out_fn)
            pmxitp.write(out_fn, scale_mass=scale_mass)

    # write hybrid topology
    print('\nlog_> Writing topology file "%s""' % outfile)
    pmxtop.write(outfile, scale_mass=scale_mass)

    # if we modified itp files included in top, fix top file
    if len(pmxitps) > 0:
        multiple_replace(outfile, replace, isfile=True)

    # separated topologies
    if args.split is True:
        write_split_top(pmxtop=pmxtop, outfile=outfile, scale_mass=scale_mass,
                        verbose=True)

    print('')
    print('b-states filled...........')
    print('')


def entry_point():
    args = parse_options()
    main(args)


if __name__ == '__main__':
    entry_point()



#!/usr/bin/env python
from pmx.estimators import BAR, Jarz, Crooks, JarzGauss
import pickle
from numpy.testing import assert_almost_equal
from numpy.random import seed


def test_BAR(gf):
    # these pickled files had to be created with py3 using protocol=2
    # in order to be py2/3 compatible
    wf = pickle.load(open(gf("dgdl/wf.pkl"), "rb"))
    wr = pickle.load(open(gf("dgdl/wr.pkl"), "rb"))

    seed(42)
    bar = BAR(wf=wf, wr=wr, T=298, nblocks=3, nboots=10)
    assert_almost_equal(bar.dg, 0.8532451987153411, decimal=7)
    assert_almost_equal(bar.err, 0.557805610661955, decimal=7)
    assert_almost_equal(bar.conv, -0.1184253367468773, decimal=7)
    assert_almost_equal(bar.err_blocks, 1.2327258529433998, decimal=7)
    assert_almost_equal(bar.err_boot, 0.4613033210812272, decimal=7)
    assert_almost_equal(bar.conv_err_boot, 0.05189927533062071, decimal=7)


def test_Jarz(gf):
    wf = pickle.load(open(gf("dgdl/wf.pkl"), "rb"))
    wr = pickle.load(open(gf("dgdl/wr.pkl"), "rb"))

    seed(42)
    jarz = Jarz(wf=wf, wr=wr, T=298, nblocks=3, nboots=10)
    assert_almost_equal(jarz.dg_mean, 0.47552516706599357, decimal=7)
    assert_almost_equal(jarz.dg_for, -0.036297126950559477, decimal=7)
    assert_almost_equal(jarz.dg_rev, 0.98734746108254667, decimal=7)
    assert_almost_equal(jarz.err_blocks_for, 1.5034007401843161, decimal=7)
    assert_almost_equal(jarz.err_blocks_rev, 0.8732303661985571, decimal=7)
    assert_almost_equal(jarz.err_boot_for, 1.31509469126237, decimal=7)
    assert_almost_equal(jarz.err_boot_rev, 0.6698096954291023, decimal=7)


def test_Crooks(gf):
    wf = pickle.load(open(gf("dgdl/wf.pkl"), "rb"))
    wr = pickle.load(open(gf("dgdl/wr.pkl"), "rb"))

    seed(42)
    cgi = Crooks(wf=wf, wr=wr, nblocks=3, nboots=10)
    assert_almost_equal(cgi.dg, 0.9312390509679932, decimal=7)
    assert_almost_equal(cgi.err_blocks, 1.4890168005551732, decimal=7)
    assert_almost_equal(cgi.err_boot1, 0.6492475842858807, decimal=7)
    assert_almost_equal(cgi.err_boot2, 0.8283881591989153, decimal=7)


def test_JarzGauss(gf):
    wf = pickle.load(open(gf("dgdl/wf.pkl"), "rb"))
    wr = pickle.load(open(gf("dgdl/wr.pkl"), "rb"))

    seed(42)
    est = JarzGauss(wf=wf, wr=wr, T=298, nblocks=3, nboots=10)
    assert_almost_equal(est.dg_for, -0.833206609449392, decimal=7)
    assert_almost_equal(est.dg_rev, 1.9406193189728818, decimal=7)
    assert_almost_equal(est.err_for, 1.4874362551150622, decimal=7)
    assert_almost_equal(est.err_rev, 1.3502363437778697, decimal=7)
    assert_almost_equal(est.err_blocks_for, 1.141666262072706, decimal=7)
    assert_almost_equal(est.err_blocks_rev, 0.9351332210549749, decimal=7)
    assert_almost_equal(est.err_boot_for, 1.9596620284612316, decimal=7)
    assert_almost_equal(est.err_boot_rev, 0.9154324171316169, decimal=7)



#!/usr/bin/env python
# pmx  Copyright Notice
# ============================
#
# The pmx source code is copyrighted, but you can freely use and
# copy it as long as you don't change or remove any of the copyright
# notices.
#
# ----------------------------------------------------------------------
# pmx is Copyright (C) 2006-2013 by Daniel Seeliger
# pmx is Copyright (C) 2013-2022 by Vytautas Gapsys
#
#                        All Rights Reserved
#
# Permission to use, copy, modify, distribute, and distribute modified
# versions of this software and its documentation for any purpose and
# without fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both the copyright notice and
# this permission notice appear in supporting documentation, and that
# the name of Daniel Seeliger not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# DANIEL SEELIGER AND VYTAUTAS GAPSYS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL DANIEL SEELIGER BE LIABLE FOR ANY
# SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ----------------------------------------------------------------------

import os
import argparse
from pmx.forcefield import *
from pmx.scripts.cli import check_unknown_cmd


def _change_outfile_format(filename, ext):
    head, tail = os.path.split(filename)
    name, ex = os.path.splitext(tail)
    new_name = os.path.join(head, name+'.'+ext)
    return new_name


# =============
# Input Options
# =============
def parse_options():
    parser = argparse.ArgumentParser(description='''
Take any itp or top and repartition mass:
triple hydrogen mass and remove the excess mass from heavy atoms
''')

    parser.add_argument('-p',
                        metavar='topol',
                        dest='intop',
                        type=str,
                        help='Input topology file (itp or top). '
                        'Default is "topol.top"',
                        default='topol.top')
    parser.add_argument('-o',
                        metavar='outfile',
                        dest='outfile',
                        type=str,
                        help='Output topology file. '
                        'Default is "pmxtop.top"',
                        default='pmxtop.top')

    args, unknown = parser.parse_known_args()
    check_unknown_cmd(unknown)

    return args


# ====
# Main
# ====
def main(args):

    top_file = args.intop
    top_file_ext = top_file.split('.')[-1]
    outfile = args.outfile

    # if input is itp but output is else, rename output
    if top_file_ext == 'itp' and outfile.split('.')[-1] != 'itp':
        outfile = _change_outfile_format(outfile, 'itp')
        print('log_> Setting outfile name to %s' % outfile)

    # load topology file
    topol = TopolBase(top_file)

    topol.repartition_mass( )

    # write topology
    topol.write( outfile )


def entry_point():
    args = parse_options()
    main(args)


if __name__ == '__main__':
    entry_point()



#!/usr/bin/env python

# pmx  Copyright Notice
# ============================
#
# The pmx source code is copyrighted, but you can freely use and
# copy it as long as you don't change or remove any of the copyright
# notices.
#
# ----------------------------------------------------------------------
# pmx is Copyright (C) 2006-2013 by Daniel Seeliger
#
#                        All Rights Reserved
#
# Permission to use, copy, modify, distribute, and distribute modified
# versions of this software and its documentation for any purpose and
# without fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both the copyright notice and
# this permission notice appear in supporting documentation, and that
# the name of Daniel Seeliger not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# DANIEL SEELIGER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL DANIEL SEELIGER BE LIABLE FOR ANY
# SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ----------------------------------------------------------------------

"""This module contains the Molecule class. It allows modifications
of structure files on the residue level.

The Molecule instance contains:

* res.atoms   (*list of atoms*)
* res.chain   (*the chain the residue belongs to*)

Examples
--------
    Basic usage:

    >>> model = Model().read(args['-f'])   # read structure file
    >>> res = model.residues[0]            # pick first residue

    Some methods:

    >>> res.set_resname('CALA')            # change residue name
    >>> a = Atom(name = 'DUM', x=[1,2,3])  # create dummy atom
    >>> res.insert_atom(0,a)               # insert atom in residue at pos 0
    >>> res.append(a)                      # append atom at the end
    >>> phi = res.get_phi()                # calculate phi angle
    >>> al = res.fetchm(['CA','C','N'])    # select multiple atoms (in that order)
    >>> res.remove_atom(al[0])             # delete atom
    >>> del res['CA']                      # delete CA atom
    >>> del res[-1]                        # delete last atom

"""

import sys
import copy
from . import library
from numpy import pi
from .atomselection import Atomselection
from .atom import Atom
from .rotamer import _aa_chi
from .geometry import Rotation
from .parser import readSection, parseList

__all__ = ['Molecule']


class Molecule(Atomselection):
    """Class for storing molecule/residue data.

    Attributes
    ----------
    resname : str
        residue/molecule name
    id : int
        residue ID
    natoms : int
        number of atoms in the Molecule
    atoms : list
        list of atoms in Molecule
    chain : Chain
        the Chain instance the Molecule is part of
    chain_id : int
        ID of the chain the residue is part of
    moltype : str
        the type of molecule/residue (protein, dna, rna, ion, water, or
        unknown).
    """

    def __init__(self, **kwargs):
        Atomselection.__init__(self)
        self.natoms = 0
        self.previous = None
        self.next = None
        self.resname = ''
        self.orig_id = 0
        self.real_resname = ''
        self.chain = None
        self.model = None
        self.chain_id = ''
        self.id = 0
        for key, val in kwargs.items():
            setattr(self, key, val)
        self.assign_moltype()

    def __str__(self):
        s = '<Molecule: moltype = %s id = %d name = %s chain_id = %s '\
            % (self.moltype, self.id, self.resname, self.chain_id)
        s += ' natoms = %d>' % len(self.atoms)
        return s

    def __getitem__(self, item):
        return self.fetch(item)[0]

    def __delitem__(self, item):
        """ delete atom """
        if item.isdigit():
            atom = self.atoms[item]
            self.remove_atom(atom)
        else:
            # assume its the atom name
            atom = self.fetch(item)[0]
            if atom:
                self.remove_atom(atom)

    def assign_moltype(self):
        """Identifies what type of molecule/residue the molecule is:
        - protein
        - dna
        - rna
        - water
        - ion
        - unknown.
        """

        # determine type
        if self.resname in library._protein_residues_incl_pmx_mut:
            self.moltype = 'protein'
        elif self.resname in library._dna_residues_incl_pmx_mut:
            self.moltype = 'dna'
        elif self.resname in library._rna_residues_incl_pmx_mut:
            self.moltype = 'rna'
        elif self.resname in library._water:
            self.moltype = 'water'
        elif self.resname in library._ions:
            self.moltype = 'ion'
        else:
            self.moltype = 'unknown'

    def is_hybrid(self):
        """Checks whether the molecule/residue is a pmx hybrid.

        Returns
        -------
        is_hybrid : bool
            whether the molecule/residue is a pmx hybrid
        """
        # check it is not an empty Molecule
        if self.resname != '':
            if (self.resname in library._perturbed_nucleotides or
               self.resname[1] == '2' or '2CM' in self.resname or
               'CM2' in self.resname):  # special two letter cases
                return True
            else:
                return False
        else:
            return False

    def has_atom(self, atom_name):
        """check whether a certain atom is present"""
        if self.fetch(atom_name):
            return True
        else:
            return False

    def new_aa(self, aa, hydrogens=True):
        """add a new amino acids (?)"""
        aa = aa.upper()
        if len(aa) == 1:
            resname = library._aacids_dic[aa]
        else:
            resname = aa
        self.resname = resname
        self.unity = 'A'
        self.atoms = []
        for i, entry in enumerate(library._aacids[resname]):
            if hydrogens is False and entry[0][0] == 'H':
                continue
            else:
                a = Atom()
                a.id = i+1
                a.name = entry[0]
                a.symbol = a.name[0]
                a.x = entry[1]
                a.occ = 1.
                a.resname = resname
                a.m = library._atommass[a.symbol]
                a.unity = 'A'
                self.atoms.append(a)
        self.set_resid(1)
        return self

    def get_real_resname(self):
        """Sets instance attribute 'real_resname'. The residue names used
        are those independent of the protonation state. E.g. for residue 'HSE',
        'HIS' is set as the 'real_resname'.
        """
        dic = {'LYP': 'LYS', 'LYSH': 'LYS', 'LYN': 'LYS', 'LSN': 'LYS',
               'CYM': 'CYS', 'CYS2': 'CYS', 'CYN': 'CYS', 'CYSH': 'CYS',
               'HIE': 'HIS', 'HIP': 'HIS', 'HID': 'HIS', 'HISA': 'HIS',
               'HISB': 'HIS', 'HSE': 'HIS', 'HSP': 'HIS',
               'HSD': 'HIS', 'HISH': 'HIS', 'HISD': 'HIS',
               'ASH': 'ASP', 'ASPP': 'ASP', 'ASPH': 'ASP',
               'GLH': 'GLU', 'GLUH': 'GLU', 'GLUP': 'GLU',
               'SEP': 'SEP', 'SEQ': 'SEQ',
               }
        if self.resname in dic:
            self.real_resname = dic[self.resname]
        else:
            self.real_resname = self.resname

    def get_psi(self, degree=False):
        """Calculate psi angle"""
        chidx = self.chain.residues.index(self)
        if chidx == len(self.chain.residues)-1:
            return -999.
        else:
            N, CA, C = self.fetchm(['N', 'CA', 'C'])
        next_mol = self.chain.residues[chidx+1]
        N2 = next_mol['N']
        dih = N.dihedral(CA, C, N2)
        if not degree:
            return dih
        else:
            return dih*180/pi

    def set_psi(self, degree, propagate=True):
        """Set psi angle"""
        psi = self.get_psi()
        diff = degree*pi/180. - psi
        CA, C = self.fetchm(['CA', 'C'])
        R = Rotation(CA.x, C.x)
        rot_atoms = self.fetch_atoms(['O', 'OXT', 'O1', 'O2', 'OC1', 'OC2'])
        if propagate:
            chidx = self.chain.residues.index(self)
            if chidx != len(self.chain.residues)-1:
                for r in self.chain.residues[chidx+1:]:
                    for atom in r.atoms:
                        rot_atoms.append(atom)
        for atom in rot_atoms:
            atom.x = R.apply(atom.x, diff)

    def set_psi_down(self, degree, propagate=True):
        psi = self.get_psi()
        diff = degree*pi/180. - psi
        CA, C = self.fetchm(['CA', 'C'])
        R = Rotation(C.x, CA.x)
        rot_atoms = self.fetch_atoms(['O', 'OXT', 'O1', 'O2',
                                      'OC1', 'OC2', 'CA', 'C'],
                                     inv=True)
        if propagate:
            chidx = self.chain.residues.index(self)
            if chidx != 0:
                for r in self.chain.residues[:chidx]:
                    for atom in r.atoms:
                        rot_atoms.append(atom)
        for atom in rot_atoms:
            atom.x = R.apply(atom.x, diff)

    def get_phi(self, degree=False):
        """calculate phi angle (C-1)-N-CA-C"""
        chidx = self.chain.residues.index(self)
        if chidx == 0:
            return -999
        else:
            previous = self.chain.residues[chidx-1]
        C = previous.fetchm(['C'])[0]
        N, CA, C2 = self.fetchm(['N', 'CA', 'C'])
        dih = C.dihedral(N, CA, C2)
        if not degree:
            return dih
        else:
            return dih*180/pi

    def set_phi(self, degree, propagate=True):
        """set phi angle (C-1)-N-CA-C"""
        if self.resname == 'PRO':
            return  # does not work
        phi = self.get_phi()
        diff = degree*pi/180. - phi
        N, CA = self.fetchm(['N', 'CA'])
        R = Rotation(N.x, CA.x)
        rot_atoms = self.fetch_atoms(['N', 'H', 'CA'], inv=True)
        if propagate:
            chidx = self.chain.residues.index(self)
            if chidx != len(self.chain.residues)-1:
                for r in self.chain.residues[chidx+1:]:
                    for atom in r.atoms:
                        rot_atoms.append(atom)
        for atom in rot_atoms:
            atom.x = R.apply(atom.x, diff)

    def set_phi_down(self, degree, propagate=True):
        if self.resname == 'PRO':
            return  # does not work
        phi = self.get_phi()
        diff = degree*pi/180. - phi
        N, CA = self.fetchm(['N', 'CA'])
        R = Rotation(CA.x, N.x)
        rot_atoms = self.fetch_atoms('H')
        if propagate:
            chidx = self.chain.residues.index(self)
            if chidx != 0:
                for r in self.chain.residues[:chidx]:
                    for atom in r.atoms:
                        rot_atoms.append(atom)
        for atom in rot_atoms:
            atom.x = R.apply(atom.x, diff)

    def get_omega(self, degree=False):
        chidx = self.chain.residues.index(self)
        if chidx == len(self.chain.residues)-1:
            return -999.
        else:
            next_mol = self.chain.residues[chidx+1]
        CA, C = self.fetchm(['CA', 'C'])
        N, CA2 = next_mol.fetchm(['N', 'CA'])
        dih = CA.dihedral(C, N, CA2)
        if not degree:
            return dih
        else:
            return dih*180/pi

    def set_omega(self, degree):
        phi = self.get_omega()
        diff = degree*pi/180. - phi
        chidx = self.chain.residues.index(self)
        if chidx == len(self.chain.residues)-1:
            return
        next_mol = self.chain.residues[chidx+1]
        C = self['C']
        N = next_mol['N']
        R = Rotation(C.x, N.x)
        rot_atoms = next_mol.fetch_atoms('N', inv=True)
        chidx = self.chain.residues.index(next_mol)
        if chidx != len(self.chain.residues)-1:
            for r in self.chain.residues[chidx+1:]:
                for atom in r.atoms:
                    rot_atoms.append(atom)
        for atom in rot_atoms:
            atom.x = R.apply(atom.x, diff)

    def set_omega_down(self, degree):
        phi = self.get_omega()
        diff = degree*pi/180. - phi
        chidx = self.chain.residues.index(self)
        if chidx == 0:
            return
        next_mol = self.chain.residues[chidx+1]
        C = self['C']
        N = next_mol['N']
        R = Rotation(N.x, C.x)
        rot_atoms = self.fetch_atoms('C', inv=True)
        chidx = self.chain.residues.index(self)
        if chidx != 0:
            for r in self.chain.residues[:chidx]:
                for atom in r.atoms:
                    rot_atoms.append(atom)
        for atom in rot_atoms:
            atom.x = R.apply(atom.x, diff)

    def nchi(self):
        self.get_real_resname()
        try:
            return len(_aa_chi[self.real_resname])
        except:
            return 0

    def get_chi(self, chi, degree=False):
        if chi > self.nchi():
            return -999
        dih_atoms = self.fetchm(_aa_chi[self.real_resname][chi][0])
        dih = dih_atoms[0].dihedral(dih_atoms[1], dih_atoms[2], dih_atoms[3])
        if not degree:
            return dih
        else:
            return dih*180./pi

    def set_chi(self, chi, phi):
        if chi > self.nchi():
            return
        ang = self.get_chi(chi)
        dih_atoms = self.fetchm(_aa_chi[self.real_resname][chi][0])
        rot_atoms = self.fetch_atoms(_aa_chi[self.real_resname][chi][1])
        delta = phi/180*pi - ang
        r = Rotation(dih_atoms[1].x, dih_atoms[2].x)
        for atom in rot_atoms:
            atom.x = r.apply(atom.x, delta)

    def set_conformation(self, rotamer):
        self.get_real_resname()
        nchi = len(_aa_chi[self.real_resname])
        for chi in range(nchi):
            self.set_chi(chi+1, rotamer[chi+1])

    def set_resname(self, resname):
        """Set the residue name.

        Parameters
        ---------
        resname : str
            the residue name
        """
        self.resname = resname
        for atom in self.atoms:
            atom.resname = resname

    def set_resid(self, resid):
        """Set the residue ID.

        Parameters
        ---------
        resname : int
            the residue index
        """
        self.id = resid
        for atom in self.atoms:
            atom.resnr = resid

    def set_orig_resid(self, resid):
        """Set the original residue/molecule ID."""
        if self.orig_id == 0:
            self.orig_id = resid

    def set_chain(self, chain):
        """Set the chain for each atom in Molecule."""
        self.chain = chain
        for atom in self.atoms:
            atom.chain = chain

    def set_molecule(self):
        """Set the molecule for each atom in Molecule."""
        for atom in self.atoms:
            atom.molecule = self

    def set_chain_id(self, chain_id):
        """Set the chain ID for each atom in Molecule."""
        self.chain_id = chain_id
        for atom in self.atoms:
            atom.chain_id = chain_id

    def insert_atom(self, pos, atom, id=True):
        """Insert an atom at a certain position.

        Parameters
        ----------
        pos : int
            index where to insert the atom
        atom : Atom
            Atom instance to insert
        id : bool, optional
            whether to assign the resnr, resname, and chain_id attributes from
            Molecule to the Atom instance. Default is True.
        """
        if pos not in range(len(self.atoms)+1):
            print('Molecule has only %d atoms' % len(self.atoms))
            return
        else:
            if id is True:
                atom.resnr = self.id
                atom.resname = self.resname
                atom.chain_id = self.chain_id
            if pos == len(self.atoms):
                at = self.atoms[-1]
                if self.model is not None:
                    idx_model = self.model.atoms.index(at)+1
                if self.chain is not None:
                    idx_chain = self.chain.atoms.index(at)+1
            else:
                at = self.atoms[pos]
                if self.model is not None:
                    idx_model = self.model.atoms.index(at)
                if self.chain is not None:
                    idx_chain = self.chain.atoms.index(at)
            atom.molecule = self
            atom.chain = self.chain
            atom.model = self.model
            if pos == len(self.atoms):
                self.atoms.append(atom)
            else:
                self.atoms.insert(pos, atom)
            if self.model is not None:
                self.model.atoms.insert(idx_model, atom)
            if self.chain is not None:
                self.chain.atoms.insert(idx_chain, atom)
            if self.model is not None:
                self.model.renumber_atoms()

    def fetch(self, key, how='byname', wildcard=False):
        """Fetch atoms. It selects atoms by name or by element.

        Parameters
        ----------
        key : str
            atom name or element
        how : str
            choose 'byname' or 'byelem'
        wildcard : bool
            whether to match the 'key' (False) or to check if the 'key' is in
            the name/element (True). Default is False.

        Returns
        -------
        result : list
            list of atoms
        """
        result = []
        if how == 'byname':
            if not wildcard:
                for atom in self.atoms:
                    if atom.name == key:
                        result.append(atom)
            else:
                for atom in self.atoms:
                    if key in atom.name:
                        result.append(atom)
        elif how == 'byelem':
            for atom in self.atoms:
                if atom.symbol == key:
                    result.append(atom)
        return result

    def fetchm(self, keys, how='byname'):
        """Fetch multiple atoms. It select atoms using a list names or
        elements.

        Parameters
        ----------
        keys : list
            list of atom names or elements
        how : str
            choose 'byname' or 'byelem'

        Returns
        -------
        result : list
            list of atoms
        """
        result = []
        if how == 'byname':
            for key in keys:
                for atom in self.atoms:
                    if atom.name == key:
                        result.append(atom)
        elif how == 'byelem':
            for atom in self.atoms:
                if atom.symbol in keys:
                    result.append(atom)
        return result

    def remove_atom(self, atom):
        """Delete an atom from Molecule.

        Parameters
        ----------
        atom : Atom
            Atom instance to be removed
        """
        if self.chain is not None:
            have_chain = True
        else:
            have_chain = False
        if self.model is not None:
            have_model = True
        else:
            have_model = False

        aidx = self.atoms.index(atom)
        if have_chain:
            chidx = self.chain.atoms.index(atom)
        if have_model:
            modidx = self.model.atoms.index(atom)
        del self.atoms[aidx]
        if have_chain:
            del self.chain.atoms[chidx]
        if have_model:
            del self.model.atoms[modidx]
            self.model.renumber_atoms()

    def append(self, atom):
        """Append an atom to the Molecule.

        Parameters
        ----------
        atom : Atom
            Atom instance to append
        """
        if not isinstance(atom, Atom):
            raise(TypeError, "%s is not an Atom instance" % str(atom))
        else:
            n = len(self.atoms)
            if n == 0:
                self.atoms.append(atom)
            else:
                self.insert_atom(n, atom)

    def copy(self):
        """ copy molecule"""
        return copy.deepcopy(self)

    # TODO: add docstring - what does this do?
    def get_bonded(self):
        bl = library._bonds[self.resname]
        for i, atom in enumerate(self.atoms[:-1]):
            for j, at in enumerate(self.atoms[i:]):
                n1 = atom.long_name.strip()
                n2 = at.long_name.strip()
                if (n1, n2) in bl or (n2, n1) in bl:
                    atom.bonds.append(at)
                    at.bonds.append(atom)

    def get_mol2_types(self, nterminus=False):
        if self.resname not in library._mol2_types:
            print('No mol2 lib entry for residue %s' % self.resname)
            sys.exit(1)
        dic = library._mol2_types[self.resname]
        for atom in self.atoms:
            if atom.symbol == 'H':
                atom.atype = 'H'
                atom.q = 0
            else:
                if atom.name in ['OC1', 'OC2', 'O1', 'O2', 'OXT']:
                    atom.atype = 'O.co2'
                    atom.q = -.5
                elif atom.name == 'N' and nterminus:
                    atom.atype = 'N.4'
                    atom.q = 1
                else:
                    atom.atype = dic[atom.name][0]
                    atom.q = dic[atom.name][1]

    # FIXME: redundant function - task can be carried out by asking whether the
    # Molecule.moltype == 'protein'
    def is_protein_residue(self):
        if self.resname in library._protein_residues:
            return True
        else:
            return False


# ==============================================================================
# SMALL MOLECULE HANDLING
# ==============================================================================

# -------
# SD file
# -------
class SDMolecule:

    def __init__(self, lst):

        self.molfile = ''
        self.properties = {}
        self.name = ''
        self.name2 = ''
        self.read(lst)

    def read(self, lst):
        self.name = lst[0].strip()
        self.name2 = lst[1].rstrip()
        s = ''
        for line in lst[2:]:
            if line.startswith('M  END'):
                s += line.rstrip()
                break
            else:
                s += line
        self.molfile = s
        idx = lst.index('M  END\n')
        for i, line in enumerate(lst[idx:]):
            if line.startswith('>'):
                prop = line.split()[1]
                self.properties[prop] = ''
                for line in lst[idx+i+1:]:
                    if line.startswith('>'):
                        break
                    else:
                        self.properties[prop] += line

    def write(self, fp):
        print(self.name, file=fp)
        print(self.name2, end='', file=fp)
        print('', file=fp)
        print(self.molfile, file=fp)
        for k, v in self.properties.items():
            print('>%s' % k, file=fp)
            print(v, end='', file=fp)
        print('$$$$', file=fp)


class SDFile:

    def __init__(self, filename=None):

        self.molecules = []
        if filename:
            self.read(filename)

    def read(self, fn):
        l = open(fn).readlines()
        lst = []
        for line in l:
            if line.startswith('$$$$'):
                mol = SDMolecule(lst)
                self.molecules.append(mol)
                lst = []
            else:
                lst.append(line)

    def write(self, fp):
        for m in self.molecules:
            m.write(fp)


# ---------
# mol2 file
# ---------
class Mol2Molecule:

    def __init__(self, lines):
        self.atom_lines = []
        self.atoms = []
        self.bonds = []
        self.bond_lines = []
        self.header = []
        self.name = ''
        self.counts = []
        self.mol_type = ''
        self.charge_type = ''
        self.footer = []
        self.keys = []
        self.__get_keys(lines)
        self.read(lines)

    def __get_keys(self, lines):
        for line in lines:
            if line.startswith('@'):
                entr = line.strip()
                if entr not in ['@<TRIPOS>MOLECULE',
                                '@<TRIPOS>ATOM',
                                '@<TRIPOS>BOND']:
                    self.keys.append(entr)

    def read(self, lines):
        self.header = readUntil(lines, '@<TRIPOS>ATOM')  # BUG? I cannot figure out where this func is coming from
        self.atom_lines = readSection(lines, '@<TRIPOS>ATOM', '@')
        self.bond_lines = readSection(lines, '@<TRIPOS>BOND', '@')
        self.__parse_molecule()
        self.__parse_atoms()
        self.__parse_bonds()
        if self.keys:
            self.footer = [self.keys[0]]+readSection(lines, self.keys[0], 'XX')

    def __parse_molecule(self):
        lines = readSection(self.header, '@<TRIPOS>MOLECULE', '@')
        self.name = lines[0].strip()
        self.counts = [int(x) for x in lines[1].strip().split()]
        self.num_atoms = self.counts[0]
        self.num_bonds = self.counts[1]
        self.num_substr = self.counts[2]
        self.num_feat = self.counts[3]
        self.num_sets = self.counts[4]

        self.mol_type = lines[2].strip()
        self.charge_type = lines[3].strip()

    def __parse_atoms(self):
        for line in self.atom_lines:
            a = Atom(mol2line=line)
            self.atoms.append(a)

    def atom_by_id(self, idx):
        for atom in self.atoms:
            if atom.id == idx:
                return atom
        return None

    def __parse_bonds(self):
        entries = parseList('iiis', self.bond_lines)
        for e in entries:
            atom1 = self.atom_by_id(e[1])
            atom2 = self.atom_by_id(e[2])
            bond_type = e[3]
            if atom1 is None or atom2 is None:
                print('Mol2Molecule: Error in bond parsing', file=sys.stderr)
                sys.exit(1)
            self.bonds.append([atom1, atom2, bond_type])

    def write(self, fp=sys.stdout):
        print('@<TRIPOS>MOLECULE', file=fp)
        print(self.name, file=fp)
        print('{0} {1} {2} {3} {4}'.format(self.num_atoms, self.num_bonds, self.num_substr, self.num_feat, self.num_sets), file=fp)
        print(self.mol_type, file=fp)
        print(self.charge_type, file=fp)
        print('', file=fp)
        print('@<TRIPOS>ATOM', file=fp)
        for atom in self.atoms:
            print("%7d %-8s %9.4f %9.4f %9.4f %-6s %3d %-8s %9.4f" %
                  (atom.id, atom.symbol, atom.x[0], atom.x[1], atom.x[2],
                   atom.atype, atom.resnr, atom.resname, atom.q), file=fp)
        print('@<TRIPOS>BOND', file=fp)
        for i, b in enumerate(self.bonds):
            print("%6d %6d %6d %6s" % ((i+1), b[0].id, b[1].id, b[2]), file=fp)
        for line in self.footer:
            print(line, file=fp)

    def add_atom(self, atom):
        n = len(self.atoms)+1
        atom.id = n
        atom.resnr = self.atoms[0].resnr
        atom.rename = self.atoms[0].resname
        self.num_atoms += 1
        self.atoms.append(atom)

    def add_bond(self, bond):
        self.num_bonds += 1
        self.bonds.append(bond)


class Mol2File:

    def __init__(self, filename=None):

        self.molecules = []
        if filename:
            self.read(filename)

    def read(self, fn):
        l = open(fn).readlines()
        for i, line in enumerate(l):
            if line.startswith('@<TRIPOS>MOLECULE'):
                lines = readUntil(l[i+1:], '@<TRIPOS>MOLECULE')  # BUG? I cannot figure out where this func is coming from
                mol = Mol2Molecule([line]+lines)
                self.molecules.append(mol)

    def write(self, fn=None):
        file_opened = False
        if fn:
            if hasattr(fn, "write"):
                fp = fn
            else:
                try:
                    fp = open(fn, 'w')
                    file_opened = True
                except:
                    fp = sys.stdout
        else:
            fp = sys.stdout
        for m in self.molecules:
            m.write(fp)
        if file_opened:
            fp.close()



#!/usr/bin/env python

# pmx  Copyright Notice
# ============================
#
# The pmx source code is copyrighted, but you can freely use and
# copy it as long as you don't change or remove any of the copyright
# notices.
#
# ----------------------------------------------------------------------
# pmx is Copyright (C) 2006-2013 by Daniel Seeliger
#
#                        All Rights Reserved
#
# Permission to use, copy, modify, distribute, and distribute modified
# versions of this software and its documentation for any purpose and
# without fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both the copyright notice and
# this permission notice appear in supporting documentation, and that
# the name of Daniel Seeliger not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# DANIEL SEELIGER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL DANIEL SEELIGER BE LIABLE FOR ANY
# SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ----------------------------------------------------------------------

"""Program to insert mutated residues in structure files for
free energy simulations.
"""

import sys
import argparse
from pmx import library
from pmx.model import Model
from pmx.parser import read_and_format
from pmx.utils import get_ff_path, ff_selection
from pmx.alchemy import mutate
from pmx.scripts.cli import check_unknown_cmd

# resinfo
dna_one_letter = {'A': 'adenosine',
                  'C': 'cytosine',
                  'G': 'guanine',
                  'T': 'thymine'}

rna_one_letter = {'A': 'adenosine',
                  'C': 'cytosine',
                  'G': 'guanine',
                  'U': 'uracil'}

# ================
# Helper functions
# ================
def _print_sorted_dict(d):
    for key in sorted(d.iterkeys()):
        print("{0:>5}{1:>15}".format(key, d[key]))


def _int_input():
    inp = input()
    try:
        inp = int(inp)
        return inp
    except:
        return( inp )
#        print('You entered "%s" -> Try again' % inp)
#        return None


def _check_residue_name(res):
    if res.resname == 'LYS':
        if res.has_atom('HZ3'):
            res.set_resname('LYP')
    elif res.resname == 'HIS':
        if res.has_atom('HD1') and \
           res.has_atom('HE2'):
            res.set_resname('HIP')
        elif res.has_atom('HD1') and not res.has_atom('HE2'):
            res.set_resname('HID')
        elif not res.has_atom('HD1') and res.has_atom('HE2'):
            res.set_resname('HIE')
    elif res.resname == 'ASP':
        if res.has_atom('HD2'):
            res.set_resname('ASH')
    elif res.resname == 'GLU':
        if res.has_atom('HE2'):
            res.set_resname('GLH')
    elif res.resname == 'CYS':
        if not res.has_atom('HG'):
            print(' Cannot mutate SS-bonded Cys %d' % res.id, file=sys.stderr)


def _ask_next():
    sys.stdout.write('\nApply another mutation [y/n]? ')
    res = input().lower()
    if res == 'y':
        return True
    elif res == 'n':
        return False
    else:
        return _ask_next()


def _match_mutation(m, ref_m, ref_chain, ref_resid):
    """Matches chain/indices of two Models. Given the chain and resid of a
    reference Model (ref_m), return the resid of the Model (m).

    Parameters
    ----------
    m: Model
        model you want to mutate
    ref_m : Model
        reference model
    ref_chain: str
        chain of residue in reference model
    ref_resid: int
        resid of residue in reference model

    Returns
    -------
    resid: int
        residue ID of model that corresponds to the chain/resid in the
        reference.
    """

    # select non-solvent residues
    res = [r for r in m.residues if r.moltype not in ['water', 'ion']]
    ref_res = [r for r in ref_m.residues if r.moltype not in ['water', 'ion']]
    # check they have same len
    assert len(res) == len(ref_res)

    # iterate through all residue pairs
    resmap = {}
    for r, rref in zip(res, ref_res):
        # first, check that the sequence is the same
        if r.resname != rref.resname:
            raise ValueError('residue %s in the input file does not match '
                             'residue %s in the input reference file'
                             % (r.resname, rref.resname))
        # then, create a dict to map (chain_id, res_id) in the reference
        # to the residues ID in the input file
        resmap[(rref.chain_id, rref.id)] = r.id

    resid = resmap[(ref_chain, ref_resid)]
    print('log_> Residue {ref_id} (chain {ref_ch}) in file {ref} mapped to residue '
          '{m_id} in file {m} after renumbering'.format(ref_ch=ref_chain,
                                                        ref_id=ref_resid,
                                                        ref=ref_m.filename,
                                                        m_id=resid,
                                                        m=m.filename))

    return resid


# ===============================
# Class for interactive selection
# ===============================
class InteractiveSelection:
    """Class containing fucntions related to the interactive selection of
    residues to be mutated.

    Parameters
    ----------
    m : Model object
        instance of pmx.model.Model
    ffpath : str
        path to forcefield files

    Attributes
    ----------
    mut_resid : int
        index of residue to be mutated
    mut_resname : str
        one-letter code of target residue

    """

    def __init__(self, m, ff, renumbered=True):
        self.m = m
        self.ffpath = get_ff_path(ff)

        # get selection
        if renumbered is True:
            self.mut_chain = None
        elif renumbered is False:
            self.mut_chain = self.select_chain()

        self.mut_resid = self.select_residue()
        self.mut_resname = self.select_mutation()

    def select_chain(self):
        """Ask for the chain id to mutate.
        """
        # show selection
        valid_ids = [c.id for c in self.m.chains]
        print('\nSelect a chain:')
        for c in self.m.chains:
            print('{0:>6}'.format(c.id))

        # select id
        selected_chain_id = None
        while selected_chain_id is None:
            sys.stdout.write('Enter chain ID: ')
            selected_chain_id = input()
            if selected_chain_id is not None and selected_chain_id not in valid_ids:
                print('Chain id %s not among valid IDs -> Try again' % selected_chain_id)
                selected_chain_id = None
        return selected_chain_id

    def select_residue(self):
        """Ask for the residue id to mutate.
        """
        # show selection if we do not need chain ID
        if self.mut_chain is None:
            valid_ids = [r.id for r in self.m.residues]
            print('\nSelect residue to mutate:')
            for i, r in enumerate(self.m.residues):
                if r.moltype not in ['water', 'ion']:
                    sys.stdout.write('%6d-%s-%s' % (r.id, r.resname, r.chain_id))
                    if (i+1) % 6 == 0:
                        print("")
        elif self.mut_chain is not None:
            valid_ids = [r.id for r in self.m.chdic[self.mut_chain].residues]
            print('\nSelect residue to mutate:')
            for i, r in enumerate(self.m.chdic[self.mut_chain].residues):
                if r.moltype not in ['water', 'ion']:
                    sys.stdout.write('%6s-%s-%s' % (str(r.id), r.resname, r.chain_id))
                    if (i+1) % 6 == 0:
                        print("")
        print("")

        # select id
        selected_residue_id = None
        while not selected_residue_id:
            sys.stdout.write('Enter residue number: ')
            selected_residue_id = _int_input()
            if (selected_residue_id is not None) and (selected_residue_id not in valid_ids):
                print('Residue id %s not among valid IDs -> Try again' % str(selected_residue_id))
                selected_residue_id = None
        return selected_residue_id

    def select_mutation(self):
        """Ask which residue to mutate to.
        """

        residue = self.m.fetch_residue(idx=self.mut_resid, chain=self.mut_chain)
        if residue.moltype == 'protein':
            aa = self.select_aa_mutation(residue)
        elif residue.moltype in ['dna', 'rna']:
            aa = self.select_nuc_mutation(residue)
        return aa

    def select_aa_mutation(self, residue):
        """Selection for protein residues.
        """

        _check_residue_name(residue)
        print('\nSelect new amino acid for %s-%s: ' % (residue.id, residue.resname))
        sys.stdout.write('Three- or one-letter code (or four-letter for ff specific residues): ')
        if residue.resname in ['HIE', 'HISE', 'HSE']:
            rol = 'X'
        elif residue.resname in ['HIP', 'HISH', 'HSP']:
            rol = 'Z'
        elif residue.resname in ['GLH', 'GLUH', 'GLUP']:
            rol = 'J'
        elif residue.resname in ['ASH', 'ASPH', 'ASPP']:
            rol = 'B'
        elif residue.resname in ['LYN', 'LYS', 'LSN']:
            rol = 'O'
        else:
            rol = library._one_letter[residue.resname]
        aa = None
        ol = list(library._aacids_dic.keys())
        tl = list(library._aacids_dic.values())
        ffpathlower = self.ffpath.lower()
        if 'amber' in ffpathlower:
                ol = list(library._aacids_ext_amber.keys())
                tl = list(library._aacids_ext_amber.values())
        if 'opls' in ffpathlower:
                ol = list(library._aacids_ext_oplsaa.keys())
                tl = list(library._aacids_ext_oplsaa.values()) + ['ASPP', 'GLUP', 'LSN']
        if 'charmm' in ffpathlower:
                ol = list(library._aacids_ext_charmm.keys())
                tl = list(library._aacids_ext_charmm.values())

        while aa is None:
            aa = input().upper()
            # some special residues:
            #   CM - deprotonated cysteine
            #   YM - deprotonated tyrosine
            if aa == 'CM':
                sys.stdout.write('Special case for deprotonated residue')
            elif len(aa) != 1 and len(aa) != 3 and len(aa) != 4:
                sys.stdout.write('Nope!\nThree- or one-letter code (or four-letter for ff specific residues): ')
                aa = None
            elif (len(aa) == 1 and aa not in ol+['B', 'J', 'O', 'X', 'Z']) or \
                 (len(aa) == 3 and aa not in tl) or \
                 (len(aa) == 4 and aa not in tl):
                sys.stdout.write('Unknown aa "%s"!\nThree- or one-letter code (or four-letter for ff specific residues): ' % aa)
                aa = None
            if aa and (len(aa) == 3 or len(aa) == 4):
                aa = library._ext_one_letter[aa]
        print('Will apply mutation %s->%s on residue %s-%s'
              % (rol, aa, residue.resname, str(residue.id)))
        return aa

    def select_nuc_mutation(self, residue):
        """Selection for nucleic acids.
        """
        aa = None
        print('\nSelect new base for %s-%s: ' % (residue.id, residue.resname))
        sys.stdout.write('One-letter code: ')
        while aa is None:
            aa = input().upper()
            if residue.moltype == 'dna' and aa not in ['A', 'C', 'G', 'T']:
                sys.stdout.write('Unknown DNA residue "%s"!\nOne-letter code: ' % aa)
                aa = None
            elif residue.moltype == 'rna' and aa not in ['A', 'C', 'G', 'U']:
                sys.stdout.write('Unknown RNA residue "%s"!\nOne-letter code: ' % aa)
                aa = None
            if aa:
                print('Will apply mutation %s->%s on residue %s-%d'
                      % (residue.resname[1], aa, residue.resname, residue.id))
            return aa


# =============
# Input Options
# =============
def parse_options():
    parser = argparse.ArgumentParser(description='''
This script applies mutations of residues in a structure file for subsequent
free energy calculations. It supports mutations to protein, DNA, and RNA
molecules.

The mutation information and dummy placements are taken from the hybrid residue
database "mutres.mtp". The best way to use this script is to take a pdb/gro file
that has been written with pdb2gmx with all hydrogen atoms present.

By default, all residues are renumbered starting from 1, so to have unique
residue IDs. If you want to keep the original residue IDs, you can use the flag
--keep_resid. In this case, you will also need to provide chain information
in order to be able to mutate the desired residue. Alternatively, if you would
like to use the original residue IDs but these have been changed, e.g. by gromacs,
you can provide a reference PDB file (with chain information too) using the --ref
flag. The input structure will be mutated according to the IDs chosen for the
reference structure after having mapped the two residue indices.

The program can either be executed interactively or via script. The script file
simply has to consist of "residue_id target_residue_name" pairs (just with some
space between the id and the name), or "chain_id residue_id target_residue_name"
if you are keeping the original residue IDs or providing a reference structure.

The script uses an extended one-letter code for amino acids to account for
different protonation states. Use the --resinfo flag to print the dictionary.

''', formatter_class=argparse.RawTextHelpFormatter)

    exclus = parser.add_mutually_exclusive_group()

    parser.add_argument('-f',
                        metavar='infile',
                        dest='infile',
                        type=str,
                        help='Input structure file in PDB or GRO format. '
                        'Default is "protein.pdb"',
                        default='protein.pdb')
    parser.add_argument('-fB',
                        metavar='infileB',
                        dest='infileB',
                        type=str,
                        help='Input structure file of the B state in PDB '
                        'or GRO format (optional).',
                        default=None)
    parser.add_argument('-o',
                        metavar='outfile',
                        dest='outfile',
                        type=str,
                        help='Output structure file in PDB or GRO format. '
                        'Default is "mutant.pdb"',
                        default='mutant.pdb')
    parser.add_argument('-ff',
                        metavar='ff',
                        dest='ff',
                        type=str.lower,
                        help='Force field to use. If none is provided, \n'
                        'a list of available ff will be shown.',
                        default=None)
    parser.add_argument('--script',
                        metavar='script',
                        dest='script',
                        type=str,
                        help='Text file with list of mutations (optional).',
                        default=None)
    exclus.add_argument('--keep_resid',
                        dest='renumber',
                        help='Whether to renumber all residues or to keep the\n'
                        'original residue IDs. By default, all residues are\n'
                        'renumbered so to have unique IDs. With this flags set,\n'
                        'the original IDs are kept. Because the IDs might not\n'
                        'be unique anymore, you will also be asked to choose\n'
                        'the chain ID where the residue you want to mutate is.',
                        default=True,
                        action='store_false')
    exclus.add_argument('--ref',
                        metavar='',
                        dest='ref_infile',
                        help='Provide a reference PDB structure from which to map\n'
                        'the chain and residue IDs onto the file to be mutated (-f).\n'
                        'This can be useful when wanting to mutate a file that\n'
                        'has had its residues renumbered or the chain information\n'
                        'removed (e.g. after gmx grompp). As in the --keep_resid\n'
                        'option, if --ref is chosen, you will need to provide chain\n'
                        'information either interactively or via the --script flag.',
                        default=None)
    parser.add_argument('--resinfo',
                        dest='resinfo',
                        help='Show the list of 3-letter -> 1-letter residues',
                        default=False,
                        action='store_true')

    args, unknown = parser.parse_known_args()
    check_unknown_cmd(unknown)

    # ------------------
    # residue dictionary
    # ------------------
    if args.resinfo is True:
        moltype = Model(args.infile).moltype
        if moltype == 'protein':
            print('\n ---------------------------')
            print(' Protein residues dictionary')
            print(' ---------------------------')
            _print_sorted_dict(library._ext_one_letter)
            print(' ---------------------------\n')
        elif moltype == 'dna':
            print('\n -----------------------')
            print(' DNA residues dictionary')
            print(' -----------------------')
            _print_sorted_dict(dna_one_letter)
            print(' -----------------------\n')
        elif moltype == 'rna':
            print('\n -----------------------')
            print(' RNA residues dictionary')
            print(' -----------------------')
            _print_sorted_dict(rna_one_letter)
            print(' -----------------------\n')
        else:
            print('\n ---------------------------')
            print(' Protein residues dictionary')
            print(' ---------------------------')
            _print_sorted_dict(library._ext_one_letter)
            print(' ---------------------------')
            print(' DNA residues dictionary')
            print(' ---------------------------')
            _print_sorted_dict(dna_one_letter)
            print(' ---------------------------')
            print(' RNA residues dictionary')
            print(' ---------------------------')
            _print_sorted_dict(rna_one_letter)
            print(' ---------------------------\n')
        exit()

    # ------------
    # ff selection
    # ------------
    if args.ff is None:
        args.ff = ff_selection()

    return args


def main(args):

    # input variables
    infile = args.infile
    infileB = args.infileB
    outfile = args.outfile
    ff = args.ff
    script = args.script
    renumber = args.renumber
    ref_infile = args.ref_infile

    # initialise Model
    m = Model(infile, renumber_residues=renumber, bPDBTER=True,
              rename_atoms=True, scale_coords='A')

    # if reference structure provided, initialise that Model too
    if ref_infile is not None:
        ref_m = Model(ref_infile, renumber_residues=False,
                      bPDBTER=True, rename_atoms=True, scale_coords='A')

    # if script is provided, do the mutations in that file
    # ----------------------------------------------------
    if script is not None:
        # 1) defualt: renumbering and not providing a ref struct
        if renumber is True and ref_infile is None:
            mutations_to_make = read_and_format(script, "is")
            # modify mut lists in mutations_to_make so that they have same
            # len of 3, both in the case where renumber is True or False
            mutations_to_make = [[None]+x for x in mutations_to_make]
        # 2) not renumbering and not providing a ref struct
        elif renumber is False and ref_infile is None:
            mutations_to_make = read_and_format(script, "sis")
        # 3) renumbering and providing a ref struct
        elif renumber is True and ref_infile is not None:
            # we read mutations according to numbering in the reference
            ref_mutations_to_make = read_and_format(script, "sis")
            mutations_to_make = []
            # we map these onto the file to be mutated
            for ref_mut in ref_mutations_to_make:
                m_resid = _match_mutation(m=m, ref_m=ref_m,
                                          ref_chain=ref_mut[0],
                                          ref_resid=ref_mut[1])
                # mut = [No Chain, resid in m, target residue]
                mut = [None, m_resid, ref_mut[2]]
                mutations_to_make.append(mut)

        # 4) NOT ALLOWED: providing a ref struct while not renumbering. This is
        # not allowed because we want to make sure there are unique indices for
        # each residue

        for mut in mutations_to_make:
            _check_residue_name(m.fetch_residue(idx=mut[1], chain=mut[0]))
            mutate(m=m,
                   mut_chain=mut[0],
                   mut_resid=mut[1],
                   mut_resname=mut[2],
                   ff=ff,
                   refB=infileB,
                   inplace=True,
                   verbose=True)

    # if script not provided, interactive selection
    # ---------------------------------------------
    else:
        do_more = True
        while do_more:
            # if no reference provided, choose from infile (Model m)
            if ref_infile is None:
                sele = InteractiveSelection(m=m, ff=ff, renumbered=renumber)
            # if reference IS provided, then choose from reference, then map
            # mutation onto infile (Model m)
            elif ref_infile is not None:
                sele = InteractiveSelection(m=ref_m, ff=ff, renumbered=False)
                sele.mut_resid = _match_mutation(m=m, ref_m=ref_m,
                                                 ref_chain=sele.mut_chain,
                                                 ref_resid=sele.mut_resid)
                sele.mut_chain = None

            # we have the needed info ==> carry out the mutation
            mutate(m=m,
                   mut_chain=sele.mut_chain,
                   mut_resid=sele.mut_resid,
                   mut_resname=sele.mut_resname,
                   ff=ff,
                   refB=infileB,
                   inplace=True,
                   verbose=True)

            # ask whether to do more mutations or stop
            if not _ask_next():
                do_more = False

    m.write(outfile)
    print('')
    print('mutations done...........')
    print('')


def entry_point():
    args = parse_options()
    main(args)


if __name__ == '__main__':
    entry_point()



#!/usr/bin/env python

# pmx  Copyright Notice
# ============================
#
# The pmx source code is copyrighted, but you can freely use and
# copy it as long as you don't change or remove any of the copyright
# notices.
#
# ----------------------------------------------------------------------
# pmx is Copyright (C) 2006-2013 by Daniel Seeliger
#
#                        All Rights Reserved
#
# Permission to use, copy, modify, distribute, and distribute modified
# versions of this software and its documentation for any purpose and
# without fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both the copyright notice and
# this permission notice appear in supporting documentation, and that
# the name of Daniel Seeliger not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# DANIEL SEELIGER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL DANIEL SEELIGER BE LIABLE FOR ANY
# SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ----------------------------------------------------------------------

"""Various utility functions and classes
"""

import os
import sys
import types
import numpy as np
import re
import logging
from glob import glob


# =============
# File IO utils
# =============

def clean_gromacs_backup_files( path ):
    """Clean gromacs backup files.

    Parameters
    ----------
    path: str
        path where to delete files starting with #
    """

    toclean = glob('{0}/*#'.format(path)) 
    for clean in toclean:
        os.remove(clean)     

def copy_files_folders( source, target ):
    """Copy files or folders.

    Parameters
    ----------
    source: str
        path to files/folders
    target: str
        target path
    """

    cmd = 'cp {0} {1} -r'.format(source,target)
    if os.path.isdir(target):
        cmd = 'cp {0} {1}/. -r'.format(source,target)
    os.system(cmd)

def create_folder( path, fname=False ):
    """Create a folder.

    Parameters
    ----------
    path: str
        full path to the folder. Alternatively, if fname is given,
        folder fname will be created in this path
    fname: str, optional
        name of the folder to create
    """ 
    if fname==False:
        if not os.path.exists(path):
            os.makedirs(path)
    elif not os.path.exists(path+'/'+fname):
        os.makedirs(path+'/'+fname)

def show_ff(gmxlib=None):
    """Prints the list of forcefields available in GMXLIB.

    Parameters
    ----------
    gmxlib : str, optional
        Path to force field library. If not set explicitly, it is taken from
        the enviroment variable GMXLIB.
    """

    if gmxlib is None:
        gmxlib = os.environ['GMXLIB']

    print('Available Force Fields in $GMXLIB:\n')
    print('  [i]    {0:40}{1}\n'.format('name', 'description'), end='')
    print('  ---    {0:40}{1}\n'.format('----', '-----------'), end='')

    ffs = [d for d in glob('{}/*.ff'.format(gmxlib)) if os.path.isdir(d)]
    ffs = natural_sort(ffs)

    for i, ff in enumerate(ffs):
        # get ff name
        f = os.path.basename(ff).split('.')[0]
        # print info
        docfile = ff + '/forcefield.doc'
        docstring = [l for l in open(docfile, 'r').readlines()][0]
        print('  [{0}]    {1:40}{2}'.format(i+1, f, docstring), end='')


def ffopen(filename, mode='r', backup=True):

    if mode == 'w':
        if os.path.isfile(filename):
            if backup:
                print('Backing up %s to %s~' % (filename, filename))
                os.rename(filename, filename+'~')
        try:
            fp = open(filename, 'w')
            return fp
        except:
            print('Error: Could not open file %s' % filename)

    elif mode == 'r':
        try:
            fp = open(filename, 'r')
            return fp
        except:
            print('No such file %s' % filename)

    else:
        return open(filename, mode)

# this function echanges netmount with home
def remove_netmount( string ): 
    regexp = re.compile('netmount')
    string = regexp.sub('home',string)
    return(string)

def get_ff_path(ff, verbose=False):
    """Get path of force field.

    Parameters
    ----------
    ff : str
        force field name

    Returns
    -------
    ff_path: str
        absolute path to the force field
    """
    ff_path = None
    # if ff is in current folder, take that path
    if os.path.isdir(ff):
        ff_path = ff
    # otherwise look into GMXLIB and GMXDATA
    else:
        # if GMXLIB is defined, look in there first
        if "GMXLIB" in os.environ:
            gmxlib = os.environ.get('GMXLIB')
            p = os.path.join(gmxlib, ff)
            pff = p+'.ff'
            if os.path.isdir(p):
                ff_path = p
            elif os.path.isdir(pff):
                ff_path = pff
        # then look into GMXDATA/top
        if "GMXDATA" in os.environ:
            gmxdata = "%s/top" % os.environ.get('GMXDATA')
            p = os.path.join(gmxdata, ff)
            pff = p+'.ff'
            if os.path.isdir(p):
                ff_path = p
            elif os.path.isdir(pff):
                ff_path = pff
        # if path still not defined, we have a problem
        if ff_path is None:
            raise ValueError('forcefield path "%s" not found' % ff)

    if verbose is True:
        print('Opening forcefield: %s' % ff_path)

    return ff_path


def get_pmxdata(fname):
    """Returns the absolute path for the file with 'fname' inside the pmx
    data folder.
    """
    pmxhome = os.path.dirname(os.path.realpath(__file__))
    pmxdata = os.path.join(pmxhome, 'data')
    filepath = '%s/%s' % (pmxdata, fname)
    if os.path.exists(filepath):
        return filepath
    else:
        raise IOError('File %s cannot be found in the following path:\n'
                      '%s' % (fname, pmxdata))



def get_mtp_file(residue, ff):
    """Returns the correct mutres mtp file depending on the nature of the
    residue.

    Parameters
    ----------
    residue : Molecule instance
        the residue that will be mutated
    ff : str
        the force field to use

    Returns
    -------
    mtp_file : str
        path to the mtp file
    """
    # Determine which mtp file to use
    ffpath = get_ff_path(ff=ff)
    # DNA mutation
    if residue.moltype == 'dna':
        mtp_file = os.path.join(ffpath, 'mutres_dna.mtp')
    # RNA mutation
    elif residue.moltype == 'rna':
        mtp_file = os.path.join(ffpath, 'mutres_rna.mtp')
    # Protein mutation
    elif residue.moltype == 'protein':
        mtp_file = os.path.join(ffpath, 'mutres.mtp')
    else:
        raise(ValueError, 'Cannot undertand mutation type needed '
                          'from the input strcture provided')
    return mtp_file


def ff_selection(gmxlib=None):
    print('Choose a force field:\n')
    print('  [i]    {0:40}{1}\n'.format('name', 'description'), end='')
    print('  ---    {0:40}{1}\n'.format('----', '-----------'), end='')

    # if no gmxlib argument passed, then get $GMXLIB
    if gmxlib is None:
        gmxlib = os.environ['GMXLIB']

    ffs = [d for d in glob('{}/*.ff'.format(gmxlib)) if os.path.isdir(d)]
    ffs = natural_sort(ffs)

    ffs_dict = {}
    for i, ff in enumerate(ffs):
        # get ff name
        f = os.path.basename(ff).split('.')[0]
        # store in dict
        ffs_dict[i+1] = f
        # print info
        docfile = ff + '/forcefield.doc'
        docstring = [l for l in open(docfile, 'r').readlines()][0]
        print('  [{0}]    {1:40}{2}'.format(i+1, f, docstring), end='')

    # allow choice of ff
    print('')
    print('Enter the index [i] of the chosen forcefield: ', end='')
    i = int(input())
    ff = ffs_dict[i]

    return ff


def listFiles(dir='./', ext=None, abs=True, backups=False):

    """ returns a list of files in
    directory dir, optionally only
    certain file types"""

    if dir[-1] != os.sep:
        dir += os.sep
    if dir[0] == '~':
        home = os.environ.get('HOME')
        dir = os.path.join(home, dir[2:])
    dir = os.path.abspath(dir)+os.sep

    l = os.listdir(dir)

    fl = []
    if not ext:
        for f in l:
            if os.path.isfile(dir+f):
                if backups:
                    if f[0] == '#' or \
                       f[-1] == '~':
                        fl.append(dir+f)
                else:
                    fl.append(dir+f)

    elif type(ext) in [types.ListType, types.TupleType]:
        for ex in ext:
            if backups:
                ff = glob(dir+'#*'+ex+'*')
                ff += glob(dir+'*.'+ex+'~')
            else:
                ff = glob(dir+'*.'+ex)
            fl.extend(ff)

    elif type(ext) == types.StringType:
        if backups:
            print(dir+'*.'+ext+'~')
            fl = glob(dir+'#*.'+ext+'*')
            fl += glob(dir+'*.'+ext+'~')
        else:
            fl.extend(glob(dir+'*.'+ext))

    if not abs:
        new = []
        for f in fl:
            new.append(f.split('/')[-1])
        return new

    return fl


def listDirs(dir='./'):
    """ returns a list of directrories in
    directory dir"""

    if dir[-1] != os.sep:
        dir += os.sep
    if dir[0] == '~':
        home = os.environ.get('HOME')
        dir = os.path.join(home, dir[2:])
    dir = os.path.abspath(dir)+os.sep

    l = os.listdir(dir)

    dl = []

    for f in l:
        if os.path.isdir(dir+f):
            dl.append(dir+f)

    return dl


def killBackups(arg, dirname, fname):

    l = listFiles(dirname, arg[0], arg[1], arg[2])
    if arg[3]:
        for f in l:
            print('%s' % f)


def removeBackups(dir, check=True):

    if dir[-1] != os.sep:
        dir += os.sep
    if dir[0] == '~':
        home = os.environ.get('HOME')
        dir = os.path.join(home, dir[2:])
    dir = os.path.abspath(dir)+os.sep

    os.path.walk(dir, killBackups, (False, True, True, check))


# ======
# Others
# ======
def doLog(fp, s, commandName='ligandHybridTop'):
    l = '{0}__log_> {1}'.format(commandName,s)
#   print(sys.stderr,l)
    print(l)
    fp.write(l+'\n')

def data2gauss(data):
    '''Takes a one dimensional array and fits a Gaussian.

    Parameters
    ----------
    data : list
        1D array of values

    Returns
    -------
    float
        mean of the distribution.
    float
        standard deviation of the distribution.
    float
        height of the curve's peak.
    '''
    m = np.average(data)
    dev = np.std(data)
    A = 1./(dev*np.sqrt(2*np.pi))
    return m, dev, A


def gauss_func(A, mean, dev, x):
    '''Given the parameters of a Gaussian and a range of the x-values, returns
    the y-values of the Gaussian function'''
    x = np.array(x)
    y = A*np.exp(-(((x-mean)**2.)/(2.0*(dev**2.))))
    return y


def natural_sort(l):
    convert = lambda text: int(text) if text.isdigit() else text.lower()
    alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)]
    return sorted(l, key=alphanum_key)


def multiple_replace(text, word_dic, isfile=False):
    """Takes a text and replace words that match a key in a dictionary with
    the associated value, returning the modified text.

    Parameters
    ----------
    text : str
        input text
    word_dic : dict
        dictionary where keys strings are replaced by their value strings
    isfile : bool, optional
        if isfile is True, the input 'text' will be considered a textfile,
        which will be opened, the patterns replaced, and written again.

    Returns
    -------
    newtext : str
        output text with strings replaced

    Examples
    --------
    >>> a = 'hello, John'
    >>> multiple_replace(a, {'hello':'goodbye', 'John':'Mark'})
    'goodbye, Mark'

    """

    rc = re.compile('|'.join(map(re.escape, word_dic)))
    def translate(match):
        return word_dic[match.group(0)]

    if isfile is False:
        return rc.sub(translate, text)
    elif isfile is True:
        f = open(text, 'r').read()
        newf = rc.sub(translate, f)
        with open(text, 'w') as out:
            out.write(newf)


def list2file(lst, fname):
    """Writes a list of strings to a file.
    """
    with open(fname, 'w') as f:
        for line in lst:
            f.write(line)


def initialise_logger(logname, logfile, level='INFO'):
    if os.path.isfile(logfile):
        os.remove(logfile)
    logger = logging.getLogger(logname)
    if level.upper() == 'INFO':
        logger.setLevel(logging.INFO)
    elif level.upper() == 'DEBUG':
        logger.setLevel(logging.DEBUG)
    else:
        raise ValueError('level can only be INFO or DEBUG')
    hdlr = logging.FileHandler(logfile)
    formatter = logging.Formatter('%(asctime)s [ %(name)s ] '
                                  '[ %(levelname)s ] %(message)s')
    hdlr.setFormatter(formatter)
    logger.addHandler(hdlr)
    return logger


def which(program):
    """Equivalent of UNIX which command. Returns the path of the executable
    if it is found, otherwise returns None.

    Parameters
    ----------
    program : str
        name of executable you are looking for

    Returns
    -------
    path : str
        string of path if found, otherwise None.

    """
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None

# ========================
# Custom Exception Classes
# ========================
class UnknownResidueError(Exception):
    """Class for unknown residue Exceptions.
    """
    def __init__(self, s):
        self.s = s

    def __str__(self):
        return repr(self.s)


class mtpError(Exception):
    """Exceptions class for ...
    """
    def __init__(self, s):
        self.s = s

    def __str__(self):
        return repr(self.s)


class MissingTopolParamError(Exception):
    """Raised when certain parameters for terms defined in the topology cannot
    ba found.
    """
    def __init__(self, s, atoms):
        self.s = s
        print('err_> {}'.format(s), file=sys.stderr)
        print('err_>  name      resname      atomtype       atomtypeB       '
              'bondtype       bondtypeB', file=sys.stderr)
        for atom in atoms:
            print('{0.name} {0.resname} {0.atomtype} {0.atomtypeB} {0.type} '
                  '{0.typeB}'.format(atom), file=sys.stderr)
        print('err_> Exiting', file=sys.stderr)

    def __str__(self):
        return repr(self.s)

class importError(Exception):
    """Exceptions class for importing a module.
    """
    def __init__(self, s):
        self.s = "Could not import: {0}".format(s)

    def __str__(self):
        return repr(self.s)




#!/usr/bin/env python

# pmx  Copyright Notice
# ============================
#
# The pmx source code is copyrighted, but you can freely use and
# copy it as long as you don't change or remove any of the copyright
# notices.
#
# ----------------------------------------------------------------------
# pmx is Copyright (C) 2006-2013 by Daniel Seeliger
#
#                        All Rights Reserved
#
# Permission to use, copy, modify, distribute, and distribute modified
# versions of this software and its documentation for any purpose and
# without fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both the copyright notice and
# this permission notice appear in supporting documentation, and that
# the name of Daniel Seeliger not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# DANIEL SEELIGER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL DANIEL SEELIGER BE LIABLE FOR ANY
# SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ----------------------------------------------------------------------

"""This module contains the Atomselection class. It contains methods
that are inherited to the Molecule, Chain and Model classes.
Take a look there for details...
"""

import sys
import random
import copy as cp
from . import library
from . import _pmx
from .atom import Atom
from .geometry import Rotation


class Atomselection:
    """ Basic class to handle sets of atoms. Atoms are stored
    in a list <atoms>"""

    def __init__(self, **kwargs):
        self.atoms = []
        self.unity = 'A'
        for key, val in kwargs.items():
            setattr(self, key, val)

    def writePDB(self, fname, title="", nr=1, bPDBTER=False,
                 bAssignChainIDs=False, resnrlist=[]):
        """Writes Atomselection to file in PDB format.

        Parameters
        ----------
        fname : str
            filename
        title : str, optional
            title of the PDB file. If not given, it is taken from the instance
            title
        nr : int, optional
            what is nr? it seems to choose whether to overwrite or append to a
            file?
        bPDBTER : bool, optional
            whether to separate chains with TER entries.
        bAssignChainIDs : bool, optional
            whether to write the chains IDs to file. Default is False.
        resnrlist : list, optional
            list of residues to write to file. If empty list provided, all
            residues are written to file, otherwise only the ones in the list
            will be written.
        """
        if nr > 1:
            fp = open(fname, 'a')
        else:
            fp = open(fname, 'w')
        if not title:
            if hasattr(self, "title"):
                title = self.title
            else:
                title = str(self.__class__)+' '+str(self)

        header = 'TITLE    '+title
        print('{}'.format(header), file=fp)
        print('MODEL%5d' % nr, file=fp)
        if not hasattr(self, "box"):
            self.box = [[0, 0, 0], [0, 0, 0], [0, 0, 0]]
        if self.box[0][0]*self.box[1][1]*self.box[2][2] != 0:
            box_line = _pmx.box_as_cryst1(self.box)
            print(box_line, file=fp)

        chainID = self.atoms[0].chain_id
        for atom in self.atoms:
            if (bPDBTER is True) and (atom.chain_id != chainID):
                print('TER', file=fp)
                chainID = atom.chain_id
            if (len(resnrlist) > 0) and (atom.resnr not in resnrlist):
                continue
            if atom.chain_id.startswith('pmx'):
                if bAssignChainIDs is False:
                    atom.chain_id = ""
                else:
                    atom.chain_id = atom.chain_id  # [-1]
            if (len(atom.name) > 4):  # too long atom name
                foo = cp.deepcopy(atom)
                foo.name = foo.name[:4]
                print(foo, file=fp)
            else:
                print(atom, file=fp)
        print('ENDMDL', file=fp)
        fp.close()

    def writeGRO(self, filename, title=''):
        fp = open(filename, 'w')
        if self.unity == 'nm':
            fac = 1.
        else:
            fac = 0.1
        if not title:
            if hasattr(self, "title"):
                title = self.title
            else:
                title = str(self.__class__)+' '+str(self)
        print(title, file=fp)
        print("%5d" % len(self.atoms), file=fp)
        if self.atoms[0].v[0] != 0.000:
            bVel = True
        else:
            bVel = False
        if bVel:
            gro_format = "%8.3f%8.3f%8.3f%8.4f%8.4f%8.4f"
        else:
            gro_format = "%8.3f%8.3f%8.3f"
        for atom in self.atoms:
            resid = (atom.resnr) % 100000
            at_id = (atom.id) % 100000
            ff = "%5d%-5.5s%5.5s%5d" % (resid, atom.resname, atom.name, at_id)
            if bVel:
                ff += gro_format % (atom.x[0]*fac,
                                    atom.x[1]*fac,
                                    atom.x[2]*fac,
                                    atom.v[0],
                                    atom.v[1],
                                    atom.v[2])
            else:
                ff += gro_format % (atom.x[0]*fac,
                                    atom.x[1]*fac,
                                    atom.x[2]*fac)
            print(ff, file=fp)

        if not hasattr(self, "box"):
            self.box = [[0, 0, 0], [0, 0, 0], [0, 0, 0]]
        if self.box[0][1] or self.box[0][2] or self.box[1][0] or \
           self.box[1][2] or self.box[2][0] or self.box[2][1]:
            bTric = False
            ff = "%10.5f%10.5f%10.5f%10.5f%10.5f%10.5f%10.5f%10.5f%10.5f"
        else:
            bTric = True
            ff = "%10.5f%10.5f%10.5f"
        if bTric:
            print(ff % (self.box[0][0], self.box[1][1], self.box[2][2]), file=fp)
        else:
            print(ff % (self.box[0][0], self.box[1][1], self.box[2][2],
                        self.box[0][1], self.box[0][2], self.box[1][0],
                        self.box[1][2], self.box[2][0], self.box[2][1]), file=fp)
        fp.close()

    def write(self, fn, title='', nr=1):
        ext = fn.split('.')[-1]
        if ext == 'pdb':
            self.writePDB(fn, title, nr)
        elif ext == 'gro':
            self.writeGRO(fn, title)
        else:
            print('pmx_Error_> Can only write pdb or gro!', file=sys.stderr)
            sys.exit(1)

    def com(self, vector_only=False):
        """move atoms to center of mass or return vector only"""
        for atom in self.atoms:
            if atom.m == 0:
                print(" Warning: Atom has zero mass: setting mass to 1.", file=sys.stderr)
                atom.m = 1.
        x = sum(map(lambda a: a.x[0]*a.m, self.atoms))
        y = sum(map(lambda a: a.x[1]*a.m, self.atoms))
        z = sum(map(lambda a: a.x[2]*a.m, self.atoms))
        M = sum(map(lambda a: a.m, self.atoms))
        x /= M
        y /= M
        z /= M
        if vector_only:
            return [x, y, z]
        else:
            for atom in self.atoms:
                atom.x[0] -= x
                atom.x[1] -= y
                atom.x[2] -= z

    def atomlistFromTop(self, topDic):
        """ return a list of atom objects
        found in topDic"""
        self.atoms = []
        for idx in topDic['atoms'].keys():
            at = Atom().atomFromTop(topDic, idx)
            self.atoms.append(at)
        return self

    def renumber_atoms(self, start=1):
        """Renumber all atoms starting from a chosen number (default is 1).
        The original atom IDs are stored in the attribute ``orig_id``

        Parameters
        ----------
        start : int, optional
            integer from which to start the indexing of the atoms
        """
        for i, atom in enumerate(self.atoms):
            if not hasattr(atom, "orig_id") or atom.orig_id == 0:
                atom.orig_id = atom.id
            atom.id = i+1

    def rename_atoms_to_gmx(self):
        """Renames atoms to comply with Gromacs syntax. If the name starts with
        a digit, the digit is moved at the end of the name. E.g. "1CA" becomes
        "CA1".
        """
        for atom in self.atoms:
            if atom.name[0].isdigit():
                atom.name = atom.name[1:]+atom.name[0]

    def a2nm(self):
        if self.unity == 'nm':
            return
        for atom in self.atoms:
            atom.x[0] *= .1
            atom.x[1] *= .1
            atom.x[2] *= .1
            atom.unity = 'nm'
        self.unity = 'nm'

    def nm2a(self):
        if self.unity == 'A':
            return
        for atom in self.atoms:
            atom.x[0] *= 10.
            atom.x[1] *= 10.
            atom.x[2] *= 10.
            atom.unity = 'A'
        self.unity = 'A'

    def get_long_name(self):
        for atom in self.atoms:
            atom.make_long_name()

    def get_symbol(self):
        for atom in self.atoms:
            atom.get_symbol()

    def get_order(self):
        for atom in self.atoms:
            atom.get_order()

    def max_crd(self):
        x = list(map(lambda a: a.x[0], self.atoms))
        y = list(map(lambda a: a.x[1], self.atoms))
        z = list(map(lambda a: a.x[2], self.atoms))
        return (min(x), max(x)), (min(y), max(y)), (min(z), max(z))

    def search_neighbors(self, cutoff=8., build_bonds=True):
        changed = False
        if self.unity == 'nm':
            changed = True  # FIXME: variable unused?
            self.nm2a()
        _pmx.search_neighbors(self.atoms, cutoff, build_bonds)

    def coords(self):
        return list(map(lambda a: a.x, self.atoms))

    def fetch_atoms(self, key, how='byname', wildcard=False, inv=False):
        '''Find atoms given their name, element, or id.

        Parameters
        ----------
        key : str|int
            key used to find the atoms.
        how : str, optional
            the way atoms are identified. Options are "byname", "byelem", or
            "byid". Default is "byname".
        wildcard : bool, optional
            whether to loosely match the atom name. In this case, if the key
            provided is within an atom name, the atom will be selected. E.g. the
            key='H1' will select atoms named, for instance, "H12", "H13", and
            "HH1".
        inv : bool, optional
            whether to invert the selection. Default is False.

        Return
        ------
        atoms: list
            list of Atom objects found.
        '''
        result = []
        if not hasattr(key, "append"):
            key = [key]
        if how == 'byname':
            for atom in self.atoms:
                for k in key:
                    if wildcard:
                        if k in atom.name:
                            result.append(atom)
                    else:
                        if atom.name == k:
                            result.append(atom)
        elif how == 'byelem':
            for atom in self.atoms:
                for k in key:
                    if atom.symbol == k:
                        result.append(atom)
        elif how == 'byid':
            for atom in self.atoms:
                for k in key:
                    if atom.id == int(k):
                        result.append(atom)
        if inv:
            r = []
            for atom in self.atoms:
                if atom not in result:
                    r.append(atom)
            return r
        else:
            return result

    def get_b13(self):
        for atom in self.atoms:
            for at2 in atom.bonds:
                for at3 in at2.bonds:
                    if atom.id > at3.id:
                        atom.b13.append(at3)
                        at3.b13.append(atom)

    def get_b14(self):
        for atom in self.atoms:
            for at2 in atom.b13:
                for at3 in at2.bonds:
                    if atom.id > at3.id and at3 not in \
                           atom.bonds+atom.b13:
                        atom.b14.append(at3)
                        at3.b14.append(atom)

    def translate(self, vec):
        for atom in self.atoms:
            atom.x[0] += vec[0]
            atom.x[1] += vec[1]
            atom.x[2] += vec[2]

    def random_rotation(self):
        vec = self.com(vector_only=True)
        self.com()
        r = Rotation([0, 0, 0], [1, 0, 0])  # x-rot
        phi = random.uniform(0., 360.)
        for atom in self.atoms:
            atom.x = r.apply(atom.x, phi)
        r = Rotation([0, 0, 0], [0, 1, 0])  # y-rot
        phi = random.uniform(0., 360.)
        for atom in self.atoms:
            atom.x = r.apply(atom.x, phi)
        r = Rotation([0, 0, 0], [0, 0, 1])  # z-rot
        phi = random.uniform(0., 360.)
        for atom in self.atoms:
            atom.x = r.apply(atom.x, phi)
        for atom in self.atoms:
            atom.x[0] += vec[0]
            atom.x[1] += vec[1]
            atom.x[2] += vec[2]

    def make_mol2_bondlist(self):
        lst = []
        bt = library._mol2_bondtypes
        for atom in self.atoms:
            for at in atom.bonds:
                if atom.id > at.id:
                    lst.append([atom, at])
        newl = []
        for at1, at2 in lst:
            check = False
            for t1, t2, tp in bt:
                if (at1.atype == t1 and at2.atype == t2) or \
                   (at1.atype == t2 and at2.atype == t1):
                    newl.append((at1, at2, tp))
                    check = True
            if not check:
                print('bondtype %s-%s defaults to 1' % (at1.atype, at2.atype))
                newl.append((at1, at2, '1'))
        self.bondlist = newl

    def get_by_id(self, lst):
        atlst = []
        for idx in lst:
            atlst.append(self.atoms[idx-1])
        return atlst



import sys
import numpy as np
import subprocess
import os
from os import path
try:
    from scipy.integrate import simps
except ImportError:
    from scipy.integrate import simpson as simps
from matplotlib import pyplot as plt
from copy import deepcopy
from scipy.special import erf
from .utils import data2gauss, gauss_func

__all__ = ['read_dgdl_files', 'integrate_dgdl',
           'ks_norm_test', 'plot_work_dist']

def find_longest_dgdl_file(lst):
    '''Takes a list of dgdl.xvg files and returns the longest one.
    The length is measured by the timestamp of the last recorded datapoint.
    
    Parameters
    ----------
    lst : list
        list containing the paths to the dgdl.xvg files.

    Returns
    -------
    longest_idx : int
        index of the longest dgdl file in lst
    t_list : list
        List of last timestamps of the dhdl files
    '''
    #find last line of every file in lst
    last_t=0.0 #length of simultaion in ps
    longest_idx=-1 #index of first full-length (longest) simultaion
    t_list=[0.0]*len(lst)
    for idx in range(len(lst)):
        if(os.path.exists(lst[idx])):
            result = subprocess.run(['tail', '-n', '1', lst[idx]],\
                                    stdout=subprocess.PIPE)
            try:
                s=result.stdout.split()
                t=float(s[0])
                t_list[idx] = t
                if t > last_t:
                    last_t = t
                    longest_idx = idx
            except:
                continue #if can't convert to float or split line
    if(longest_idx<0):
        raise RuntimeError("No valid dgdl file found.")
        
    return(longest_idx, t_list)


def read_dgdl_files(lst, lambda0=0, invert_values=False, verbose=True,\
                    sigmoid=0.0):
    '''Takes a list of dgdl.xvg files and returns the integrated work values.

    Parameters
    ----------
    lst : list
        list containing the paths to the dgdl.xvg files.
    lambda0 : [0,1]
        whether the simulations started from lambda 0 or 1. Default is 0.
    invert_values : bool
        whether to invert the sign of the returned work value.

    Returns
    -------
    w : array
        array of work values.
    lst : list
        sorted list of input dgdl.avg files corresponding to the work values
        in w.
    '''

    # check lambda0 is either 0 or 1
    assert lambda0 in [0, 1]
    
    #Start with first full-length (longest) simultaion.
    #Everything before is too short.
    good=False
    idx, t_list=find_longest_dgdl_file(lst) 
    last_t=t_list[idx]
    first_w=0
    ndata=0
    while(not good and idx<len(lst)):
        if(t_list[idx] >= last_t): #only check full length simulations
            try:
                _check_dgdl(lst[idx], lambda0, verbose=verbose)
                first_w, ndata = integrate_dgdl(lst[idx], lambda0=lambda0,\
                                            invert_values=invert_values,\
                                            sigmoid=sigmoid)
                good=True
            except:
                print(' !! Error in checking %s' % (lst[idx]))
                good=False
                idx+=1
        else:
            idx+=1
    if(not good):
        raise RuntimeError("No good dgdl files provided.")

    w_list = [first_w]
    for idx, f in enumerate(lst[idx+1:]):
        if verbose is True:
            sys.stdout.write('\r    Reading %s' % f)
            sys.stdout.flush()

        w, _ = integrate_dgdl(f, ndata=ndata, lambda0=lambda0,
                              invert_values=invert_values, sigmoid=sigmoid)
        if w is not None:
            w_list.append(w)

    if verbose is True:
        print('\n')
    return w_list


def integrate_dgdl(fn, ndata=-1, lambda0=0, invert_values=False, sigmoid=0.0):
    '''Integrates the data in a dgdl.xvg file.

    Parameters
    ----------
    fn : str
        the inpur dgdl.xvg file from Gromacs.
    ndata : int, optional
        number of datapoints in file. If -1, then ??? default is -1.
    lambda0 : [0,1]
        whether the simulations started from lambda 0 or 1. Default is 0.
    invert_values : bool
        whether to invert the sign of the returned work value.

    Returns
    -------
    integr : float
        result of the integration performed using Simpson's rule.
    ndata : int
        number of data points in the input file.
    '''

    # check lambda0 is either 0 or 1
    assert lambda0 in [0, 1]

    try:
        lines = open(fn, encoding="ISO-8859-1").readlines()
    except:
        return None, None
    if not lines:
        return None, None

    # extract dgdl datapoints into r
    # TODO: we removed the check for file integrity. We could have an
    # optional files integrity check before calling this integration func

    lines = [l for l in lines if l[0] not in '#@&']
    r=[]
    try:
        r = list(map(lambda x: float(x.split()[1]), lines))
    except:
        print(' !! Error in reading %s' % (fn))
        return None, None

    if ndata != -1 and len(r) != ndata:
        try:
            print(' !! Skipping %s ( read %d data points, should be %d )'
                  % (fn, len(r), ndata))
        except:
            print(' !! Skipping %s ' % (fn))
        return None, None
    # convert time to lambda
    ndata = len(r)
    dlambda = 1./float(ndata)
    if lambda0 == 1:
        dlambda *= -1

    # arrays for the integration
    # --------------------------
    # array of lambda values
    x = [lambda0+i*dlambda for i, dgdl in enumerate(r)]
##### VG VG VG ####
    if sigmoid != 0.0:
        x = [1.0/(1.0+np.exp( -1.0*(lambda0+i*dlambda-0.5)*sigmoid ) ) for i, dgdl in enumerate(r)]
        x[0] = lambda0
        x[-1] = 1.0-lambda0
#    for i in x:
#        print(i)
##### VG VG VG ####
    # array of dgdl
    y = r

    if lambda0 == 1:
        x.reverse()
        y.reverse()

    if invert_values is True:
        integr = simps(y, x) * (-1)
        return integr, ndata
    else:
        integr = simps(y, x)
        return integr, ndata


def ks_norm_test(data, alpha=0.05, refks=None):
    '''Performs a Kolmogorov-Smirnov test of normality.

    Parameters
    ----------
    data : array_like
        a one-dimensional array of values. This is the distribution tested
        for normality.
    alpha : float
        significance level of the statistics. Default if 0.05.
    refks : ???
        ???

    Returns
    -------
    Q : float
    lam0 : float
    check : float
    bOk : bool
    '''

    def ksref():
        f = 1
        potent = 10000
        lamb = np.arange(0.25, 2.5, 0.001)
        q = np.zeros(len(lamb), float)
        res = []
        for k in range(-potent, potent):
            q = q + f*np.exp(-2.0*(k**2)*(lamb**2))
            f = -f
        for i in range(len(lamb)):
            res.append((lamb[i], q[i]))
        return res

    def ksfunc(lamb):
        f = 1
        potent = 10000
        q = 0
        for k in range(-potent, potent):
            q = q + f*np.exp(-2.0*(k**2)*(lamb**2))
            f *= -1
        return q

    def edf(dg_data):
        edf_ = []
        ndata = []
        data = deepcopy(dg_data)
        data.sort()
        N = float(len(data))
        cnt = 0
        for item in data:
            cnt += 1
            edf_.append(cnt/N)
            ndata.append(item)
        ndata = np.array(ndata)
        edf_ = np.array(edf_)
        return ndata, edf_

    def cdf(dg_data):
        data = deepcopy(dg_data)
        data.sort()
        mean = np.average(data)
        sig = np.std(data)
        cdf = 0.5*(1+erf((data-mean)/float(sig*np.sqrt(2))))
        return cdf

    N = len(data)
    nd, ed = edf(data)
    cd = cdf(data)
    siglev = 1-alpha
    dval = []
    for i, val in enumerate(ed):
        d = abs(val-cd[i])
        dval.append(d)
        if i:
            d = abs(ed[i-1]-cd[i])
            dval.append(d)
    dmax = max(dval)
    check = np.sqrt(N)*dmax
    if not refks:
        refks = ksref()
    lst = list(filter(lambda x: x[1] > siglev, refks))
    lam0 = lst[0][0]
    if check >= lam0:
        bOk = False
    else:
        bOk = True

    q = ksfunc(check)
    return (1-q), lam0, check, bOk


def plot_work_dist(wf=[], wr=[], fname='Wdist.png', nbins=20, dG=None, dGerr=None,
                   units='kJ/mol', dpi=300, statesProvided='AB'):
    '''Plots forward and reverse work distributions. Optionally, it adds the
    estimate of the free energy change and its uncertainty on the plot.

    Parameters
    ----------
    wf : list
        list of forward work values.
    wr : list
        list of reverse work values.
    fname : str, optional
        filename of the saved image. Default is 'Wdist.png'.
    nbins : int, optional
        number of bins to use for the histogram. Default is 20.
    dG : float, optional
        free energy estimate.
    dGerr : float, optional
        uncertainty of the free energy estimate.
    units : str, optional
        the units of dG and dGerr. Default is 'kJ/mol'.
    dpi : int
        resolution of the saved image file.
    statesProvided: str
        work values for two states or only one

    Returns
    -------
    None

    '''

    def smooth(x, window_len=11, window='hanning'):

        if x.ndim != 1:
            raise ValueError("smooth only accepts 1 dimension arrays.")
        if x.size < window_len:
            raise ValueError("Input vector needs to be bigger than "
                             "window size.")
        if window_len < 3:
            return x
        if window not in ['flat', 'hanning', 'hamming',
                          'bartlett', 'blackman']:
            raise ValueError("Window is on of 'flat', 'hanning', 'hamming', "
                             "'bartlett', 'blackman'")
        s = np.r_[2*x[0]-x[window_len:1:-1], x, 2*x[-1]-x[-1:-window_len:-1]]
        # moving average
        if window == 'flat':
            w = np.ones(window_len, 'd')
        else:
            w = eval('np.' + window + '(window_len)')
        y = np.convolve(w/w.sum(), s, mode='same')
        return y[window_len-1:-window_len+1]

    plt.figure(figsize=(8, 6))
    x1 = list(range(len(wf)))
    x2 = list(range(len(wr)))
    if 'A' in statesProvided:
        x1 = list(range(len(wf)))
    if 'B' in statesProvided:
        x2 = list(range(len(wr)))
    if len(x1) > len(x2):
        x = x1
    else:
        x = x2
    if 'A' in statesProvided:
        mf, devf, Af = data2gauss(wf)
    if 'B' in statesProvided:
        mb, devb, Ab = data2gauss(wr)

    if 'AB' in statesProvided: 
        maxi = max(wf+wr)
        mini = min(wf+wr)
        plt.subplot(1, 2, 1)
        plt.plot(x1, wf, 'g-', linewidth=2, label="Forward (0->1)", alpha=.3)
        plt.plot(x2, wr, 'b-', linewidth=2, label="Backward (1->0)", alpha=.3)
        ### smoothing ###
        try:
            sm1 = smooth(np.array(wf))
            plt.plot(x1, sm1, 'g-', linewidth=3)
        except:
            print("Plotting: no smoothing for Wf")
        try:
            sm2 = smooth(np.array(wr))
            plt.plot(x2, sm2, 'b-', linewidth=3)
        except:
            print("Plotting: no smoothing for Wr")
    elif 'A' in statesProvided: 
        maxi = max(wf)
        mini = min(wf)
        plt.subplot(1, 2, 1)
        plt.plot(x1, wf, 'g-', linewidth=2, label="Forward (0->1)", alpha=.3)
        ### smoothing ###
        try:
            sm1 = smooth(np.array(wf))
            plt.plot(x1, sm1, 'g-', linewidth=3)
        except:
            print("Plotting: no smoothing for Wf")
    elif 'B' in statesProvided: 
        maxi = max(wr)
        mini = min(wr)
        plt.subplot(1, 2, 1)
        plt.plot(x2, wr, 'b-', linewidth=2, label="Backward (1->0)", alpha=.3)
        ### smoothing ###
        try:
            sm2 = smooth(np.array(wr))
            plt.plot(x2, sm2, 'b-', linewidth=3)
        except:
            print("Plotting: no smoothing for Wr")

    plt.legend(shadow=True, fancybox=True, loc='upper center',
               prop={'size': 12})
    plt.ylabel(r'W [kJ/mol]', fontsize=20)
    plt.xlabel(r'# Snapshot', fontsize=20)
    plt.grid(lw=2)
    plt.xlim(0, x[-1]+1)
    xl = plt.gca()
    for val in xl.spines.values():
        val.set_lw(2)
    plt.subplot(1, 2, 2)
    if 'A' in statesProvided:
        plt.hist(wf, bins=nbins, orientation='horizontal', facecolor='green',
             alpha=.75, density=True)
    if 'B' in statesProvided:
        plt.hist(wr, bins=nbins, orientation='horizontal', facecolor='blue',
             alpha=.75, density=True)

    x = np.arange(mini, maxi, .5)

    if 'AB' in statesProvided:
        y1 = gauss_func(Af, mf, devf, x)
        y2 = gauss_func(Ab, mb, devb, x)
        plt.plot(y1, x, 'g--', linewidth=2)
        plt.plot(y2, x, 'b--', linewidth=2)
        size = max([max(y1), max(y2)])
    elif 'A' in statesProvided:
        y1 = gauss_func(Af, mf, devf, x)
        plt.plot(y1, x, 'g--', linewidth=2)
        size = max(y1)
    elif 'B' in statesProvided:
        y2 = gauss_func(Ab, mb, devb, x)
        plt.plot(y2, x, 'b--', linewidth=2)
        size = max(y2)

    res_x = [dG, dG]
    res_y = [0, size*1.2]
    if dG is not None and dGerr is not None:
        plt.plot(res_y, res_x, 'k--', linewidth=2,
                 label=r'$\Delta$G = %.2f $\pm$ %.2f %s' % (dG, dGerr, units))
        plt.legend(shadow=True, fancybox=True, loc='upper center',
                   prop={'size': 12})
    elif dG is not None and dGerr is None:
        plt.plot(res_y, res_x, 'k--', linewidth=2,
                 label=r'$\Delta$G = %.2f %s' % (dG, units))
        plt.legend(shadow=True, fancybox=True, loc='upper center',
                   prop={'size': 12})
    else:
        plt.plot(res_y, res_x, 'k--', linewidth=2)

    plt.xticks([])
    plt.yticks([])
    xl = plt.gca()
    for val in xl.spines.values():
        val.set_lw(2)
    plt.subplots_adjust(wspace=0.0, hspace=0.1)
    plt.savefig(fname, dpi=dpi)


def _check_dgdl(fn, lambda0, verbose=True):
    '''Prints some info about a dgdl.xvg file.'''
    lines = open(fn, encoding="ISO-8859-1").readlines()
    if not lines:
        return None
    r = []
    for line in lines:
        if line[0] not in '#@&':
            r.append([float(x) for x in line.split()])
    ndata = len(r)
    dlambda = 1./float(ndata)
    if lambda0 == 1:
        dlambda *= -1

    if verbose is True:
        print('    First accepted file: %s' % fn)
        print('    # data points: %d' % ndata)
        print('    Length of trajectory: %8.3f ps' % r[-1][0])
        print('    Delta lambda: %8.5f' % dlambda)



#!/usr/bin/env python
# pmx  Copyright Notice
# ============================
#
# The pmx source code is copyrighted, but you can freely use and
# copy it as long as you don't change or remove any of the copyright
# notices.
#
# ----------------------------------------------------------------------
# pmx is Copyright (C) 2006-2013 by Daniel Seeliger
#
#                        All Rights Reserved
#
# Permission to use, copy, modify, distribute, and distribute modified
# versions of this software and its documentation for any purpose and
# without fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both the copyright notice and
# this permission notice appear in supporting documentation, and that
# the name of Daniel Seeliger not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# DANIEL SEELIGER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL DANIEL SEELIGER BE LIABLE FOR ANY
# SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ----------------------------------------------------------------------

import sys,os
from pmx import *
from pmx.options import *
from pmx.parser import *
from pmx import library
from pmx.mutdb import *
from pmx.geometry import *

amber_termini_full = {
    'NALA':'ALA',
    'NGLY':'GLY',
    'NSER':'SER',
    'NTHR':'THR',
    'NLEU':'LEU',
    'NILE':'ILE',
    'NVAL':'VAL',
    'NASN':'ASN',
    'NGLN':'GLN',
    'NARG':'ARG',
    'NHID':'HID',
    'NHIE':'HIE',
    'NHIP':'HIP',
    'NTRP':'TRP',
    'NPHE':'PHE',
    'NTYR':'TYR',
    'NGLU':'GLU',
    'NASP':'ASP',
    'NLYP':'LYP',
    'NPRO':'PRO',
    'NCYS':'CYS',
    'NCYN':'CYN',
    'NCYX':'CYX',
    'NMET':'MET',
    'CALA':'ALA',
    'CGLY':'GLY',
    'CSER':'SER',
    'CTHR':'THR',
    'CLEU':'LEU',
    'CILE':'ILE',
    'CVAL':'VAL',
    'CASN':'ASN',
    'CGLN':'GLN',
    'CARG':'ARG',
    'CHID':'HID',
    'CHIE':'HIE',
    'CHIP':'HIP',
    'CTRP':'TRP',
    'CPHE':'PHE',
    'CTYR':'TYR',
    'CGLU':'GLU',
    'CASP':'ASP',
    'CLYP':'LYP',
    'CPRO':'PRO',
    'CCYS':'CYS',
    'CCYN':'CYN',
    'CCYX':'CYX',
    'CMET':'MET'
}

amber_termini_partial = {
    'NAL':'ALA',
    'NSE':'SER',
    'NTH':'THR',
    'NLE':'LEU',
    'NIL':'ILE',
    'NVA':'VAL',
    'NAR':'ARG',
    'NHI':'HIS',
    'NTR':'TRP',
    'NPH':'PHE',
    'NTY':'TYR',
    'NLY':'LYS',
    'NPR':'PRO',
    'NCY':'CYS',
    'CAL':'ALA',
    'CSE':'SER',
    'CTH':'THR',
    'CLE':'LEU',
    'CIL':'ILE',
    'CVA':'VAL',
    'CAR':'ARG',
    'CHI':'HIS',
    'CTR':'TRP',
    'CPH':'PHE',
    'CTY':'TYR',
    'CLY':'LYS',
    'CPR':'PRO',
    'CCY':'CYS',
    'CME':'MET',
}

def identify_nas_cas(r):
    newname = 'ASP'
    for a in r.atoms:
        if a.name=='ND2':
            return('ASN')
        if a.name=='OD2':
            return('ASP')
    return(newname)

def identify_gly_glu_gln(r):
    newname = 'GLY'
    for a in r.atoms:
        if a.name=='OE2':
            return('GLU')
        if a.name=='NE2':
            return('GLN')
    return(newname)

def identify_nmet_nme(r):
    newname = 'NME'
    for a in r.atoms:
        if 'S' in a.name:
            return('MET')
    return(newname)

def fix_amber_termini(m,ff):
    # get the termini
    termini = []
    for ch in m.chains:
        termini.append(ch.residues[0])
        termini.append(ch.residues[-1])

    # check (and fix) the names
    dict_full = amber_termini_full.keys()
    dict_part = amber_termini_partial.keys()
    for term in termini:
	# terminal name in the 4 letter dictionary
        if term.resname in dict_full:
            term.set_resname(amber_termini_full[term.resname])
            continue
	# terminal name in the 3 letter dictionary
        if term.resname in dict_part:
            term.set_resname(amber_termini_partial[term.resname])
            continue
	# other special cases
        if term.resname.startswith('NAS') or term.resname.startswith('CAS'):
            newname = identify_nas_cas(term)
            term.set_resname(newname)
            continue
        if term.resname.startswith('NCY') or term.resname.startswith('CCY'):
            newname = 'CYS'
            term.set_resname(newname)
            continue
        if term.resname.startswith('NGL') or term.resname.startswith('CGL'):
            newname = identify_gly_glu_gln(term)
            term.set_resname(newname)
            continue
        if term.resname.startswith('NHI') or term.resname.startswith('CHI'):
            newname = 'HIS'
            term.set_resname(newname)
            continue
        if term.resname.startswith('NLY') or term.resname.startswith('CLY'):
            newname = 'LYS'
            term.set_resname(newname)
            continue
        if term.resname.startswith('NME'):
            newname = identify_nmet_nme(term)
            term.set_resname(newname)
            continue

def rename_lowlevel(r,name):
    for a in r.atoms:
        a.resname = name

def rename_hie(r,ff):
    if ff.startswith('charmm'):
        newname = 'HSE'
    if ff.startswith('amber'):
        newname = 'HIE'
    if ff.startswith('opls'):
        newname = 'HISE'
    # actual rename
    rename_lowlevel(r,newname)

def rename_hid(r,ff):
    if ff.startswith('charmm'):
        newname = 'HSD'
    if ff.startswith('amber'):
        newname = 'HID'
    if ff.startswith('opls'):
        newname = 'HISD'
    # actual rename
    rename_lowlevel(r,newname)

def rename_hip(r,ff):
    if ff.startswith('charmm'):
        newname = 'HSP'
    if ff.startswith('amber'):
        newname = 'HIP'
    if ff.startswith('opls'):
        newname = 'HISH'
    # actual rename
    rename_lowlevel(r,newname)

def rename_his(r,ff):
    # identify histidine type
    bNd = False
    bNe = False
    for a in r.atoms:
        if a.name == 'HE2':
            bNe = True
        if a.name == 'HD1':
            bNd = True

    # no Hydrogen names
#    newname = 'HIE'
#    if ff.startswith('charmm'):
#	newname = 'HSE'
#    elif ff.startswith('opls'):
#	newname = 'HISE'

    # names with Hydrogens
    if (bNd==True) and (bNe==True):
        if ff.startswith('charmm'):
            newname = 'HSP'
        if ff.startswith('amber'):
            newname = 'HIP'
        if ff.startswith('opls'):
            newname = 'HISH'
        # actual rename
        rename_lowlevel(r,newname)
    elif (bNd==False) and (bNe==True):
        if ff.startswith('charmm'):
            newname = 'HSE'
        if ff.startswith('amber'):
            newname = 'HIE'
        if ff.startswith('opls'):
            newname = 'HISE'
        # actual rename
        rename_lowlevel(r,newname)
    elif (bNd==True) and (bNe==False):
        if ff.startswith('charmm'):
            newname = 'HSD'
        if ff.startswith('amber'):
            newname = 'HID'
        if ff.startswith('opls'):
            newname = 'HISD'
        # actual rename
        rename_lowlevel(r,newname)

def rename_lys(r,ff):
    # default names
    newname = 'LYS'
    if ff.startswith('amber') or ff.startswith('charmm'):
        newname = 'LYS'
    elif ff.startswith('opls'):
        newname = 'LYSH'

    # names with hydrogens
    bUncharged = False
    for a in r.atoms:
        if a.name == 'HZ2':
            bUncharged = True
        if a.name == 'HZ3':
            bUncharged = False
            break

    # if lysine neutral
    if bUncharged==True:
        if ff.startswith('amber'):
            newname = 'LYN'
        elif ff.startswith('charmm'):
            newname = 'LSN'
        elif ff.startswith('opls'):
            newname = 'LSN'

    # actual rename
    rename_lowlevel(r,newname)

def rename_asp(r,ff):
    # default names
    newname = 'ASP'

    # names with hydrogens
    bUncharged = False
    for a in r.atoms:
        if a.name == 'HD2':
            bUncharged = True
            break

    # if aspartate neutral
    if bUncharged==True:
        if ff.startswith('amber'):
            newname = 'ASH'
        elif ff.startswith('charmm'):
            newname = 'ASPP'
        elif ff.startswith('opls'):
            newname = 'ASPP'

    # actual rename
    rename_lowlevel(r,newname)

def rename_glu(r,ff):
    # default names
    newname = 'GLU'

    # names with hydrogens
    bUncharged = False
    for a in r.atoms:
        if a.name == 'HE2':
            bUncharged = True
            break

    # if glutamate neutral
    if bUncharged==True:
        if ff.startswith('amber'):
            newname = 'GLH'
        elif ff.startswith('charmm'):
            newname = 'GLUP'
        elif ff.startswith('opls'):
            newname = 'GLUP'

    # actual rename
    rename_lowlevel(r,newname)

def rename_cys(r,ff):
    # default names
    newname = 'CYS'
    if ff.startswith('opls'):
        newname = 'CYSH'

    # names with hydrogens
    bDisulfide = False
    for a in r.atoms:
        if a.name == 'HB1':
            bDisulfide = True
        if a.name.startswith('HG'):
            bDisulfide = False
            break

    # if disulfide
    if bDisulfide==True:
        if ff.startswith('amber'):
            newname = 'CYX'
        elif ff.startswith('charmm'):
            newname = 'CYS2'
        elif ff.startswith('opls'):
            newname = 'CYS2'

    # actual rename
    rename_lowlevel(r,newname)

def rename_cyx(r,ff):
    if ff.startswith('amber'):
        newname = 'CYX'
    elif ff.startswith('charmm'):
        newname = 'CYS2'
    elif ff.startswith('opls'):
        newname = 'CYS2'
    # actual rename
    rename_lowlevel(r,newname)

def rename_lsn(r,ff):
    if ff.startswith('amber'):
        newname = 'LYN'
    elif ff.startswith('charmm'):
        newname = 'LSN'
    elif ff.startswith('opls'):
        newname = 'LSN'
    # actual rename
    rename_lowlevel(r,newname)

def rename_aspp(r,ff):
    if ff.startswith('amber'):
        newname = 'ASH'
    elif ff.startswith('charmm'):
        newname = 'ASPP'
    elif ff.startswith('opls'):
        newname = 'ASPP'
    # actual rename
    rename_lowlevel(r,newname)

def rename_glup(r,ff):
    if ff.startswith('amber'):
        newname = 'GLH'
    elif ff.startswith('charmm'):
        newname = 'GLUP'
    elif ff.startswith('opls'):
        newname = 'GLUP'
    # actual rename
    rename_lowlevel(r,newname)

def rename_res(m,ff):
    for r in m.residues:
        if r.resname == 'HIS':
            rename_his(r,ff)
        if r.resname == 'HIE' or r.resname == 'HISE' or r.resname == 'HSE':
            rename_hie(r,ff)
        if r.resname == 'HID' or r.resname == 'HISD' or r.resname == 'HSD':
            rename_hid(r,ff)
        if r.resname == 'HIP' or r.resname == 'HISH' or r.resname == 'HSP':
            rename_hip(r,ff)
        if r.resname.startswith('LYS') or r.resname == 'LYP':
            rename_lys(r,ff)
        if r.resname == 'LSN' or r.resname == 'LYN':
            rename_lsn(r,ff)
        if r.resname == 'ASP':
            rename_asp(r,ff)
        if r.resname == 'ASH' or r.resname == 'ASPP':
            rename_aspp(r,ff)
        if r.resname =='GLU':
            rename_glu(r,ff)
        if r.resname == 'GLH' or r.resname == 'GLUP':
            rename_glup(r,ff)
        if r.resname == 'CYS' or r.resname == 'CYSH' or r.resname == 'CYN':
            rename_cys(r,ff)
        if r.resname == 'CYX' or r.resname == 'CYS2':
            rename_cyx(r,ff)

def rename_caps(m,ff):
# Amber: NME and ACE
# Charmm: CT3 and ACE
# OPLS: NAC and ACE
    for r in m.residues:
        if r.resname=='CT3': # charmm
            if ff.startswith('amber'):
                rename_lowlevel(r,'NME')
            elif ff.startswith('opls'):
                rename_lowlevel(r,'NAC')
            elif ff.startswith('maestro'):
                rename_lowlevel(r,'NMA')
        if r.resname=='NAC': # opls
            if ff.startswith('amber'):
                rename_lowlevel(r,'NME')
            elif ff.startswith('charmm'):
                rename_lowlevel(r,'CT3')
            elif ff.startswith('maestro'):
                rename_lowlevel(r,'NMA')
        if r.resname=='NME':
            if ff.startswith('opls'):
                rename_lowlevel(r,'NAC')
            elif ff.startswith('charmm'):
                rename_lowlevel(r,'CT3')
        if r.resname=='NMA': # maestro
            if ff.startswith('opls'):
                rename_lowlevel(r,'NAC')
            elif ff.startswith('charmm'):
                rename_lowlevel(r,'CT3')
            elif ff.startswith('amber'):
                rename_lowlevel(r,'NME')

def terminal_oxt(m):
    for a in m.atoms:
        if a.name=='OXT':
            a.name = 'O'
    
def main(argv):

   options=[
   Option( "-ft", "string", "charmm" , "force field type (charmm, amber99sb, amber99sb*-ildn, oplsaa"),
#   Option( "-align", "bool", True, "align side chains"),
        ]
    
   files = [
       FileOption("-f", "r",["pdb","gro"],"protein.pdb", "input structure file"),
       FileOption("-o", "w",["pdb","gro"],"out.pdb", "output structure file"),
       ]
    
   help_text = ('If hydrogens in the system are to be preserved, ',
                'this script would adjust residue names',
		'for pdb2gmx not to complain.',
		'Residues considered: His, Lys, Asp, Glu.',
                'Also the termini names will be fixed.'
                '',
                )

    
   cmdl = Commandline( argv, options = options,
                       fileoptions = files,
                       program_desc = help_text,
                       check_for_existing_files = False )
    
   iname = cmdl['-f']
   oname = cmdl['-o']
   ff = cmdl['-ft']

   m = Model(iname,bPDBTER=True,renumber_residues=False)

   # fix potential problems with the amber termini
   fix_amber_termini(m,ff)

   # rename residues properly
   rename_res(m,ff)

   # fix terminal OXT, if present
   terminal_oxt(m)

   # rename capping groups
   rename_caps(m,ff)

   m.write(oname,bPDBTER=True)

if __name__=='__main__':
    main(sys.argv)




#!/usr/bin/env python

# pmx  Copyright Notice
# ============================
#
# The pmx source code is copyrighted, but you can freely use and
# copy it as long as you don't change or remove any of the copyright
# notices.
#
# ----------------------------------------------------------------------
# pmx is Copyright (C) 2006-2013 by Daniel Seeliger
#
#                        All Rights Reserved
#
# Permission to use, copy, modify, distribute, and distribute modified
# versions of this software and its documentation for any purpose and
# without fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both the copyright notice and
# this permission notice appear in supporting documentation, and that
# the name of Daniel Seeliger not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# DANIEL SEELIGER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL DANIEL SEELIGER BE LIABLE FOR ANY
# SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ----------------------------------------------------------------------

"""Module containing fuctions dealing with Gromacs index files.
"""

import re
import sys,os,random
from .parser import readSection
from .model import Model


class IndexGroup:
    """Class for storing information about a Gromacs index group.

    Parameters
    ----------
    name : str
        name of the group
    ids : list, optional
        list of atom indices
    atoms : list, optional
        list of ``Atom`` instances. If this is provided, the ids will be taken from
        the Atom instances.

    Attributes
    ----------
    name : str
        name of the group
    ids : list
        list of atom indices
    """
    def __init__(self, name='', ids=[], atoms=[]):
        self.ids = []
        if atoms:
            self.ids = list(map(lambda a: a.id, atoms))
        else:
            self.ids = ids
        self.name = name
        self.ids.sort()

    def __str__(self):
        s = ''
        s += '[ %s ]\n' % self.name
        count = 0
        for i in range(len(self.ids)):
            if len(str(self.ids[i])) > 6:
                fs = "%d "
            else:
                fs = "%6d "
            s += fs % self.ids[i]
            count += 1
            if count % 15 == 0:
                s += '\n'
        return s

    def read_index_group(self, name, lines):
        if '[' in name:
            name = name[1:-1].strip()
        self.name = name
        key = '[ '+name+' ]'
        ll = readSection(lines, key, '[')
        self.ids = []
        for line in ll:
            self.ids.extend([int(x) for x in line.split()])
        return self

    def select_atoms(self, atom_sel):
        atoms = []
        for idx in self.ids:
            for atom in atom_sel.atoms:
                if atom.id == idx:
                    atoms.append(atom)
        return atoms


class IndexFile:
    """Class for storing ``ndx`` file information. It can be initialsed as an empty
    index file, or by providing an ``ndx`` file as input, or by providing a list
    of names and ``IndexGroup`` instances.

    Parameters
    ----------
    fname : str, optional
        input index file. Default is None.
    names : list, optional
        list of the group names.
    groups : list, optional
        list of IndexGroup instances.

    Attributes
    ----------
    groups : list
        list of IndexGroup instances
    names : list
        list if index group names
    dic : dict
        dictionary containg index group names as keys and ``IndexGroup``
        instances as values
    """

    def __init__(self, fname=None, names=[], groups=[]):
        self.groups = []
        self.names = []
        self.dic = {}
        if fname is not None:
            self.parse(fname)
        if groups:
            for g in groups:
                self.add_group(g)

    def __str__(self):
        s = 'Gromacs index file (%d index groups):\n' % len(self.groups)
        for i, group in enumerate(self.groups):
            s += ' %d  %-20s   :    %d atoms\n' % (i, group.name, len(group.ids))
        return s

    def __getitem__(self, item):
        return self.dic[item]

    def __delitem__(self, item):
        self.delete_group(item)

    def parse(self, fp):
        """Reads an index file.

        Parameters
        ----------
        fp : str|object
            filename of the index file, or the file object of the ndx file
            already opened.
        """
        if hasattr(fp, "read"):
            f = fp.read()
        else:
            f = open(fp).read()
        names = self.__get_index_groups(f)
        lines = f.split('\n')
        for name in names:
            idx = IndexGroup().read_index_group(name, lines)
            self.add_group(idx)
        return self

    def __get_index_groups(self, f):
        x = re.compile('\[ .* \]\n')
        names = x.findall(f)
        r = []
        for name in names:
            r.append(name.strip())
        return r

    def write(self, fn=None, fp=None):
        """Writes the index file.

        Parameters
        ----------
        fn : str, optional
            filename. Default is ``None``. If both ``fn`` and ``fp`` are
            ``None``, the index is printed to screen.
        fp : object, optional
            file object to write the index to. Default is ``None``.
        """
        if not fn and not fp:
            fp = sys.stdout
        if fn:
            fp = open(fn, 'w')
        for gr in self.groups:
            print('{0}\n'.format(str(gr)), file=fp)
        if fn:
            fp.close()

    def add_group(self, group):
        """Adds a group to the IndexFile.

        Parameters
        ----------
        group : IndexGroup
            instance of IndexGroup to add
        """
        if group.name in self.names:
            print("IndexFile has group %s !! " % group.name, file=sys.stderr)
            print("Group %s will be replaced !!" % group.name, file=sys.stderr)
            self.delete_group(group.name)
        self.names.append(group.name)
        self.groups.append(group)
        self.dic[group.name] = group

    def delete_group(self, name):
        """Removes a group from the IndexFile.

        Parameters
        ----------
        name : str
            name of the group to remove
        """
        idx = -1
        for i, group in enumerate(self.groups):
            if group.name == name:
                idx = i
                break
        if idx == -1:
            return
        del self.groups[idx]
        del self.dic[name]
        self.names.remove(name)


def get_index(atom_list=None, residue_list=None, chain_list=None):
    """ return atom indices from a list of atoms/residues/chains"""
    if not atom_list and not residue_list and not chain_list:
        print('Error: Need list~')
        sys.exit(1)
    if atom_list:
        lst = list(map(lambda a: a.id, atom_list))
        return lst
    if residue_list:
        al = []
        list(map(lambda r: al.extend(r.atoms), residue_list))
        return get_index(atom_list=al)
    if chain_list:
        al = []
        list(map(lambda c: al.extend(c.atoms), chain_list))
        return get_index(atom_list=al)


def make_index_group(atomlist, name):
    lst = get_index(atomlist)
    g = IndexGroup(ids=lst, name=name)
    return g


def create_ndx_freeze( pdbFile, ndxFile='index_FREEZE.ndx', groupname='FREEZE' ):
    """Create a group for freezing with all atoms except dummies.

    Parameters
    ----------
    pdbfile : str
            filename of pdb. The group will be constructed based on the atom names.
    ndxFile : str, optional
            filename of ndx. An existing or not existing index file can be provided.
    groupname : str, optional
            name for the new group
    """

    # deal with the index file
    if os.path.isfile(ndxFile): # file exists
        indFile = IndexFile(ndxFile)
        randnum = random.randint(1000,9999)
        if groupname in indFile.names:
            groupname = '{0}{1}'.format(groupname,randnum)
    else: # new file
        indFile = IndexFile()
    
    # create index group
    m = Model(pdbFile)
    i = 0
    ind = []
    for a in m.atoms:
        i+=1
        if a.name.startswith('D') or a.name.startswith('HV'):
            continue
        ind.append(i)
    indGroup = IndexGroup(name=groupname,ids=ind)
    
    # add index group to the file
    indFile.add_group(indGroup)
    
    # write
    indFile.write(fn=ndxFile)
    




import sys
from pmx import library
from pmx.utils import get_ff_path, ff_selection

# ===============================
# Class for interactive selection
# ===============================
class InteractiveSelection:
    """Class containing functions related to the interactive selection of
    residues to be mutated.

    Parameters
    ----------
    m : Model object
        instance of pmx.model.Model
    ffpath : str
        path to forcefield files

    Attributes
    ----------
    mut_resid : int
        index of residue to be mutated
    mut_resname : str
        one-letter code of target residue

    """

    def __init__(self, m, ff=None, renumbered=True):
        self.m = m
        self.ffpath = ''
        if ff==None:
            self.ffpath = get_ff_path(ff)

        # get selection
        if renumbered is True:
            self.mut_chain = None
        elif renumbered is False:
            self.mut_chain = self.select_chain()

        self.mut_resid = self.select_residue()
        self.mut_resname = self.select_mutation()

    def select_chain(self):
        """Ask for the chain id to mutate.
        """
        # show selection
        valid_ids = [c.id for c in self.m.chains]
        print('\nSelect a chain:')
        for c in self.m.chains:
            print('{0:>6}'.format(c.id))

        # select id
        selected_chain_id = None
        while selected_chain_id is None:
            sys.stdout.write('Enter chain ID: ')
            selected_chain_id = input()
            if selected_chain_id is not None and selected_chain_id not in valid_ids:
                print('Chain id %s not among valid IDs -> Try again' % selected_chain_id)
                selected_chain_id = None
        return selected_chain_id

    def select_residue(self):
        """Ask for the residue id to mutate.
        """
        # show selection if we do not need chain ID
        if self.mut_chain is None:
            valid_ids = [r.id for r in self.m.residues]
            print('\nSelect residue to mutate:')
            for i, r in enumerate(self.m.residues):
                if r.moltype not in ['water', 'ion']:
                    sys.stdout.write('%6d-%s-%s' % (r.id, r.resname, r.chain_id))
                    if (i+1) % 6 == 0:
                        print("")
        elif self.mut_chain is not None:
#            valid_ids = [r.id for r in self.m.chdic[self.mut_chain].residues]
            valid_ids = []
            for r in self.m.chdic[self.mut_chain].residues:
                if isinstance(r.id,str):
                    valid_ids.append(r.id.replace(" ",""))
                else:
                    valid_ids.append(r.id)
            print('\nSelect residue to mutate:')
            for i, r in enumerate(self.m.chdic[self.mut_chain].residues):
                if r.moltype not in ['water', 'ion']:
                    sys.stdout.write('%6s-%s-%s' % (r.id, r.resname, r.chain_id))
                    if (i+1) % 6 == 0:
                        print("")
        print("")

        # select id
        selected_residue_id = None
        while not selected_residue_id:
            sys.stdout.write('Enter residue number: ')
            selected_residue_id = _int_input()
            if selected_residue_id is not None and selected_residue_id not in valid_ids:
                print('Residue id %s not among valid IDs -> Try again' % selected_residue_id)
                selected_residue_id = None
        return selected_residue_id

    def select_mutation(self):
        """Ask which residue to mutate to.
        """

        residue = self.m.fetch_residue(idx=self.mut_resid, chain=self.mut_chain)
        if residue.moltype == 'protein':
            aa = self.select_aa_mutation(residue)
        elif residue.moltype in ['dna', 'rna']:
            aa = self.select_nuc_mutation(residue)
        return aa

    def select_aa_mutation(self, residue):
        """Selection for protein residues.
        """

        _check_residue_name(residue)
        print('\nSelect new amino acid for %s-%s: ' % (residue.id, residue.resname))
        sys.stdout.write('Three- or one-letter code (or four-letter for ff specific residues): ')
        if residue.resname in ['HIE', 'HISE', 'HSE']:
            rol = 'X'
        elif residue.resname in ['HIP', 'HISH', 'HSP']:
            rol = 'Z'
        elif residue.resname in ['GLH', 'GLUH', 'GLUP']:
            rol = 'J'
        elif residue.resname in ['ASH', 'ASPH', 'ASPP']:
            rol = 'B'
        elif residue.resname in ['LYN', 'LYS', 'LSN']:
            rol = 'O'
        else:
            rol = library._one_letter[residue.resname]
        aa = None
        ol = list(library._aacids_dic.keys())
        tl = list(library._aacids_dic.values())
        ffpathlower = self.ffpath.lower()
        if 'amber' in ffpathlower:
                ol = list(library._aacids_ext_amber.keys())
                tl = list(library._aacids_ext_amber.values())
        if 'opls' in ffpathlower:
                ol = list(library._aacids_ext_oplsaa.keys())
                tl = list(library._aacids_ext_oplsaa.values()) + ['ASPP', 'GLUP', 'LSN']
        if 'charmm' in ffpathlower:
                ol = list(library._aacids_ext_charmm.keys())
                tl = list(library._aacids_ext_charmm.values())

        while aa is None:
            aa = input().upper()
            # some special residues:
            #   CM - deprotonated cysteine
            #   YM - deprotonated tyrosine
            if aa == 'CM':
                sys.stdout.write('Special case for deprotonated residue')
            elif len(aa) != 1 and len(aa) != 3 and len(aa) != 4:
                sys.stdout.write('Nope!\nThree- or one-letter code (or four-letter for ff specific residues): ')
                aa = None
            elif (len(aa) == 1 and aa not in ol+['B', 'J', 'O', 'X', 'Z']) or \
                 (len(aa) == 3 and aa not in tl) or \
                 (len(aa) == 4 and aa not in tl):
                sys.stdout.write('Unknown aa "%s"!\nThree- or one-letter code (or four-letter for ff specific residues): ' % aa)
                aa = None
            if aa and (len(aa) == 3 or len(aa) == 4):
                aa = library._ext_one_letter[aa]
        print('Will apply mutation %s->%s on residue %s-%s'
              % (rol, aa, residue.resname, residue.id))
        return aa

    def select_nuc_mutation(self, residue):
        """Selection for nucleic acids.
        """
        aa = None
        print('\nSelect new base for %s-%s: ' % (residue.id, residue.resname))
        sys.stdout.write('One-letter code: ')
        while aa is None:
            aa = input().upper()
            if residue.moltype == 'dna' and aa not in ['A', 'C', 'G', 'T']:
                sys.stdout.write('Unknown DNA residue "%s"!\nOne-letter code: ' % aa)
                aa = None
            elif residue.moltype == 'rna' and aa not in ['A', 'C', 'G', 'U']:
                sys.stdout.write('Unknown RNA residue "%s"!\nOne-letter code: ' % aa)
                aa = None
            if aa:
                print('Will apply mutation %s->%s on residue %s-%d'
                      % (residue.resname[1], aa, residue.resname, residue.id))
            return aa

# ===============================
# Helper functions
# ===============================
def _int_input():
    inp = input()
    try:
        inp = int(inp)
        return inp
    except:
        print('You entered "%s" -> Will try to proceed assuming this is a valid ID' % inp)
        return inp
#        return None

def _check_residue_name(res):
    if res.resname == 'LYS':
        if res.has_atom('HZ3'):
            res.set_resname('LYP')
    elif res.resname == 'HIS':
        if res.has_atom('HD1') and \
           res.has_atom('HE2'):
            res.set_resname('HIP')
        elif res.has_atom('HD1') and not res.has_atom('HE2'):
            res.set_resname('HID')
        elif not res.has_atom('HD1') and res.has_atom('HE2'):
            res.set_resname('HIE')
    elif res.resname == 'ASP':
        if res.has_atom('HD2'):
            res.set_resname('ASH')
    elif res.resname == 'GLU':
        if res.has_atom('HE2'):
            res.set_resname('GLH')
    elif res.resname == 'CYS':
        if not res.has_atom('HG'):
            print(' Cannot mutate SS-bonded Cys %d' % res.id, file=sys.stderr)


def _match_mutation(m, ref_m, ref_chain, ref_resid):
    """Matches chain/indices of two Models. Given the chain and resid of a
    reference Model (ref_m), return the resid of the Model (m).

    Parameters
    ----------
    m: Model
        model you want to mutate
    ref_m : Model
        reference model
    ref_chain: str
        chain of residue in reference model
    ref_resid: int
        resid of residue in reference model

    Returns
    -------
    resid: int
        residue ID of model that corresponds to the chain/resid in the
        reference.
    """

    # select non-solvent residues
    res = [r for r in m.residues if r.moltype not in ['water', 'ion']]
    ref_res = [r for r in ref_m.residues if r.moltype not in ['water', 'ion']]
    # check they have same len
    assert len(res) == len(ref_res)

    # iterate through all residue pairs
    resmap = {}
    for r, rref in zip(res, ref_res):
        # first, check that the sequence is the same
        if r.resname != rref.resname:
            raise ValueError('residue %s in the input file does not match '
                             'residue %s in the input reference file'
                             % (r.resname, rref.resname))
        # then, create a dict to map (chain_id, res_id) in the reference
        # to the residues ID in the input file
        resmap[(rref.chain_id, rref.id)] = r.id

    resid = resmap[(ref_chain, ref_resid)]
    print('log_> Residue {ref_id} (chain {ref_ch}) in file {ref} mapped to residue '
          '{m_id} in file {m} after renumbering'.format(ref_ch=ref_chain,
                                                        ref_id=ref_resid,
                                                        ref=ref_m.filename,
                                                        m_id=resid,
                                                        m=m.filename))

    return resid
def _match_mutation(m, ref_m, ref_chain, ref_resid):
    """Matches chain/indices of two Models. Given the chain and resid of a
    reference Model (ref_m), return the resid of the Model (m).

    Parameters
    ----------
    m: Model
        model you want to mutate
    ref_m : Model
        reference model
    ref_chain: str
        chain of residue in reference model
    ref_resid: int
        resid of residue in reference model

    Returns
    -------
    resid: int
        residue ID of model that corresponds to the chain/resid in the
        reference.
    """

    # select non-solvent residues
    res = [r for r in m.residues if r.moltype not in ['water', 'ion']]
    ref_res = [r for r in ref_m.residues if r.moltype not in ['water', 'ion']]
    # check they have same len
    assert len(res) == len(ref_res)

    # iterate through all residue pairs
    resmap = {}
    for r, rref in zip(res, ref_res):
        # first, check that the sequence is the same
        if r.resname != rref.resname:
            raise ValueError('residue %s in the input file does not match '
                             'residue %s in the input reference file'
                             % (r.resname, rref.resname))
        # then, create a dict to map (chain_id, res_id) in the reference
        # to the residues ID in the input file
        resmap[(rref.chain_id, rref.id)] = r.id

    resid = resmap[(ref_chain, ref_resid)]
    print('log_> Residue {ref_id} (chain {ref_ch}) in file {ref} mapped to residue '
          '{m_id} in file {m} after renumbering'.format(ref_ch=ref_chain,
                                                        ref_id=ref_resid,
                                                        ref=ref_m.filename,
                                                        m_id=resid,
                                                        m=m.filename))

    return resid
